#!/bin/env python3
# =====================================================================
''' PROGRAM INFORMATION
Program:  LGRS_Grid_Generation_mk6.py
Language: Python 3.11.4
Author:   Mark T. McClernan
Created:  March, 2024   (Python 3.11.4)
Modified: April, 2024   (Python 3.11.4)
IDE:      Visual Studio Code (v1.18.1 x86)

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
PROGRAM DESCRIPTION:
    This program generates ESRI shapefiles to display the following 
    lunar projected coordinate reference systems (PCRS) or grid systems: 
        1. Lunar Transverse Mercator (LTM);
        2. Lunar Polar Stereoscopic (LPS);
        3. Lunar Grid Reference System (LGRS);
    and
        4. LGRS in Artemis Condensed Coordinate Format (ACC)
        5. LGRS ACC coordinates limited to 6-characters.
    Because of the condensed format of LGRS in ACC format, when specifying
    input regions only 1 - 4 is able to be specified, LGRS in ACC format 
    assumes the knowledge of the LGRS 25km grid location. The grid generation 
    program will need this information to geospatially reference the correct 
    location.  Grids are generated by specifying an input and output coordinate 
    format and a bounding box that dimensions the grid area in formats 1 -4. 
    
    Some grids only require that a portion in the grid area be specified, these
    are generally small-scale and often cover a global resolution. For 
    simplicity, this software only creates shapefiles. If the bounding box 
    
    covers more than one area, the grid will be generated for both areas and
    joined. For these areas covering multiple zones, the files will be
    projected in LatLon regardless of which zones the grid area is plotted
    in. This software makes use of the LGRS_Coordinate_Conversion_mk6.py
    program for all grid conversions and utilizes the OSGeo: GDAL, OGR,
    and OGR python packages for reading and writing shapefiles.

    This software makes use of the LGRS_Coordinate_Conversion_mk3.py 
    program for all grid conversions and utilizes the OSGeo: GDAL, OGR,
    and OGR python API packages for reading and writing shapefiles.

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
HOW TO USE THIS PROGRAM

    This software is designed to run via the command line, where an input
    and output format are provided with the grid size and input region or 
    coordinate box. The program will create and esri shapefile or geopack
    file with the grid.

    The basic way to call the program is similar to calling the Grid 
    Generation program is similar to the Coordinate Conversion program in
    in put structure with some difference. 
    
    The program is called from the command line with the sample command:

    ./LGRS_Grid_Generation_mk6.py {Input format} {Output Format} \
        {Grid resolution} {Location or (ULX ULY LRX LRY)}

    Only a specific combination of the input format, output format, and
    the desired grid will produce an output file. Not all grid resolutions
    are available, depending on the input or output format. The program has built
    in warnings and checks that a correct output will be only be generated
    if the correct inputs are specified. 
    
    Program Required argument descriptions

    {Input format}
        The accepted input formats are 
        LatLon, LTM, LPS, LGRS, PolarLGRS, LGRS_ACC, and PolarLGRS_ACC.
        ACC or PolarACC cannot be specified on their own as these are relative 
        coordinates. 

        The input formats are:
        •	LatLon – Lat lon pair between latitudes -82°<=lat<=+-82°
        •	PolarLatLon – polar Lat lon pair between latitudes <-80° or >+80°
        •	LTM – Lunar Transverse Mercator
        •	LPS – Lunar Polar Stereographic 
        •	LGRS – Lunar Gird Reference System
        •	LGRS_ACC - Lunar Gird Reference System in Artemis Condensed Format
        •	PolarLGRS - Lunar Gird Reference System in polar regions
        •	PolarLGRS_ACC - Lunar Gird Reference System in Artemis Condensed 
        Format in polar regionsSpecifying this format will change the way the 
        input coordinates are read in to match input format

    {Output Format}
        Acceptable output coordinates are 
        LatLon, LTM, LPS, LGRS, PolarLGRS, LGRS_ACC, PolarLGRS_ACC, 
        ACC or PolarACC.

        Output formats are the same and include two additional grids. 
        •	LatLon
        •	PolarLatLon
        •	LTM
        •	LPS
        •	LGRS
        •	LGRS_ACC
        •	PolarLGRS
        •	PolarLGRS_ACC
        •	ACC - Lunar Gird Reference System in Artemis Condensed Format
                truncated to 6 characters. 
        •	PolarACC - Lunar Gird Reference System in Artemis Condensed 
                    Format in polar regions truncated to 6 characters.
        The output format only changes the way the grid areas are named, 
        the grid positioning will be in LTM or LPS. 

    {Input format} -> {Output Format}
        This program does not provide the native ability to conduct inverse 
        coordinate conversions like the LGRS_Coordonate_Conversion_mk* 
        code. As such, only forward conversions from the input 
        coordinate to the output coordinate will produce a grid shapefile. 

        List of acceptable I/O form in form out combinations:
        •	Form in     -> Form Out
        -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  
        •	LatLon          -> LTM, LGRS, LGRS_ACC, ACC
        •	PolarLatLon     -> LPS, PolarLGRS, PolarLGRS_ACC, PolarACCPolarACC
        •	LTM             -> LTM, LGRS, LGRS_ACC, ACC
        •	LPS             -> LPS, PolarLGRS, PolarLGRS_ACC, PolarACC
        •	LGRS            -> LGRS, LGRS_ACC, ACC
        •	PolarLGRS       -> PolarLGRS, PolarLGRS_ACC, PolarACC
        •	LGRS_ACC        -> LGRS_ACC, ACC
        •	PolarLGRS_ACC   -> PolarLGRS_ACC, PolarACC
        -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  
        Inverse conversion should be completed with the Coordinate Conversion 
        Program

    {Grid resolution} 
        This program is capable of writing shapefiles for all grid resolutions
        of LTM, LPS, LGRS, and ACC formatted grids for the lunar equatorial 
        and polar regions. There are some naming conventions 

        Types of grids and descriptions
        coarse grid {   •	global - Generate grid at global scale.
                                    Used for LPS or LTM zone boarder in LPS/LTM coordinates
                                    LGRS Global area reference (8°x8°) or (10°x8°) grid in LTM zone
                        •	global_all - Generate all grids global scale grids if available
                                    LTM zone boarders and LGRS Global areas
                                    All LGRS (8°x8°) or (10°x8°) global areas
                                    LGRS Polar border, LPS boarder, for one hemisphere
                        •	25km - Generate 25km topocentric grid within a single zone
                                    Generate LGRS 25km grid within one LTM zone
                                    Generate Polar LGRS 25km grid within one LPS zone
                        •	25km_all - Generate plot of all 25km grid zones. 
                                        Generate all 25km grid zones within all LTM zones.
                                        Generate all 25km grid zones within one LPS zone.
                    }
        fine grid { •	1km     - 1km LGRS grid in LGRS or ACC format
                    •	100m    - 100m LGRS grid in LGRS or ACC format
                    •	10m     - 10m LGRS grid in LGRS or ACC format
                    •	1m      - 1m LGRS grid in LGRS or ACC format
                }
    {Location or (ULX ULY LRX LRY)}
        Specifying the plotting region of the grid has two formats, 
        
        "Coarse" grids, defined as grid areas > 25km, only require the 
        plotting region. This is either the LTM zone ("Zh") of the LPS
        Zone ("h").

        "Fine" grids, defines as grids areas < 25km, require input
        of a bounding box to restrict the plotting region. To input
        these arguments, specific the upper left (UL) and lower 
        right (LR) corners as follows ULX ULY LRX LRY, where
        ULX - UL X or Easting
        ULY - UL Y or Northing
        LRX - LR X or Easting
        LRY - LR Y or Northing
        There arguments can be specified in planetocentric lat/lon
        or topocentric Eastings and Northings. It found to be easier 
        to specify the grid area with topocentric coordinates during 
        program development and testing.

        For grid areas that cross multiple LTM zones, the user will 
        be prompted whether they would like a multigrid. If yes, a
        grid will be created that crosses multiple zones. To support 
        this operation, multigrid coordinates will only be exported in
        planetocentric Lat/Lon. Program will terminate if a multigrid
        is not a desired output. This option is not needed for grids
        within a single LTM zone. Additionally, multigrids are not 
        available to plot between the LTM and LPS portions of LPS, as 
        the LPS and LTM portions are a notable boundary of LGRS and 
        a multigrid is not really needed in polar regions. For data 
        near the +-80° LPS/LGRS boarder, use the LTM extended range.

    Example program calls
        Coarse Grids
            LatLon
                ./LGRS_Grid_Generation_mk6.py LatLon LTM border 30 45
                ./LGRS_Grid_Generation_mk6.py LatLon LTM border -30 -45

            PolarLatLon
                ./LGRS_Grid_Generation_mk6.py PolarLatLon LPS global 0 -85
                ./LGRS_Grid_Generation_mk6.py PolarLatLon LPS global 0 -85

            LTM
                ./LGRS_Grid_Generation_mk6.py LTM LTM border ${i} N

            LPS
                ./LGRS_Grid_Generation_mk6.py PolarLatLon LPS global 0 -85
                ./LGRS_Grid_Generation_mk6.py LPS LPS border N

            LGRS
                ./LGRS_Grid_Generation_mk6.py LTM LGRS global_all
                ./LGRS_Grid_Generation_mk6.py LTM LGRS 25km_all

                ./LGRS_Grid_Generation_mk6.py LTM LGRS global 23N
                ./LGRS_Grid_Generation_mk6.py LTM LGRS 25km 23N

            PolarLGRS
                ./LGRS_Grid_Generation_mk6.py LPS PolarLGRS global N
                ./LGRS_Grid_Generation_mk6.py LPS PolarLGRS global_all S

                ./LGRS_Grid_Generation_mk6.py LPS PolarLGRS 25km S
                ./LGRS_Grid_Generation_mk6.py LPS PolarLGRS 25km N

        fine grids
            equatorial LGRS_ACC (full name)
            ./LGRS_Grid_Generation_mk6.py LGRS_ACC LGRS_ACC 1km 30PED-000-000 30PFC-000-000
            ./LGRS_Grid_Generation_mk6.py LGRS_ACC LGRS_ACC 100m 30PED-000A000 30PEDA000-000
            ./LGRS_Grid_Generation_mk6.py LGRS_ACC LGRS_ACC 10m 30PED-000-100 30PED-100-000
            ./LGRS_Grid_Generation_mk6.py LGRS_ACC LGRS_ACC 1m 30PED-000-100 30PED-100-000

            polar PolarLGRS_ACC (full name)
            ./LGRS_Grid_Generation_mk6.py LPS PolarLGRS_ACC 1km S475000E600000N S525000E550000N
            ./LGRS_Grid_Generation_mk6.py LPS PolarLGRS_ACC 100m S475000E600000N S525000E550000N
            ./LGRS_Grid_Generation_mk6.py LPS PolarLGRS_ACC 10m S495000E575000N S505000E565000N
            ./LGRS_Grid_Generation_mk6.py LPS PolarLGRS_ACC 1m S495000E575000N S505000E565000N
   
        multigrid
            ./LGRS_Grid_Generation_mk6.py LatLon LGRS 1km 42.5 1.5 45.5 -1.5 
            ./LGRS_Grid_Generation_mk6.py LatLon LGRS 100m 43.5 0.5 44.5 -0.5
            ./LGRS_Grid_Generation_mk6.py LatLon LGRS 10m 43.75 0.25 44.25 -0.25
            ./LGRS_Grid_Generation_mk6.py LatLon LGRS 1m 43.95 0.05 44.05 -0.05

    Artemis III Leibnitz Beta Plateau Nobile Rim 1 
    Candidate Landing sites -  ACC grid example (truncated)

        Coarse grids
            ./LGRS_Grid_Generation_mk6.py LPS PolarLGRS global S
            ./LGRS_Grid_Generation_mk6.py LPS PolarLGRS 25km S
        fine grids
            ./LGRS_Grid_Generation_mk6.py PolarLatLon PolarACC 1km 27.288174340696408 -85.3270828326219 43.22853025996592 -85.47632419407181
            ./LGRS_Grid_Generation_mk6.py PolarLatLon PolarACC 100m 27.288174340696408 -85.3270828326219 43.22853025996592 -85.47632419407181
        WARNING: The following command produce extremely large files and take some time to process. 
            ./LGRS_Grid_Generation_mk6.py PolarLatLon PolarACC 10m 27.288174340696408 -85.3270828326219 43.22853025996592 -85.47632419407181
            1m not available in ACC

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
PROGRAM SECTIONS (can search for each using "SECTION #"):
    SECTION 0: INITIALIZE PROGRAM AND CALL EXTERNAL LIBRARIES
    SECTION 1: REFERENCE SURFACE PARAMETERS
        SECTION 1.1: LIST COORDINATE FORMATS
        SECTION 1.2: DEFINE MAP PROJECTIONS FOR PCRS AND GRID SYSTEMS USING
        SECTION 1.3: Define the LTM LPS coordinate limits and resolution
    SECTION 2: FUNCTION LIBRARY
        SECTION 2.1: Naming attribute table when creating shape files
        SECTION 2.2: Assign values to attribute table
        SECTION 2.3: Export multi polygon grid to esri shape-file
        SECTION 2.4: Export single polygon, as a multi polygon data type, to
        SECTION 2.5: Assign X Y Coordinates to multi polygon osgeo ogr
        SECTION 2.6: Assign X Y positional Coordinates to single ogr ring
        SECTION 2.7: Assign X Y positional Coordinates to ogr rings in multi
                    polygon geometry
        SECTION 2.8: Intersect ogr multi polygon geometries to a single geometry
        SECTION 2.9: Merge ogr multi polygon geometries to a single geometry
        SECTION 2.10: Clip multi polygon geometries within another geometry
        SECTION 2.11: Extract X, Y positions from multi polygon geometry
        SECTION 2.12: Calculate hemisphere location on a sphere in degrees
        SECTION 2.13: calculate LTM zone prime meridian from a LTM zone number
        SECTION 2.14: Determine the LTM zone number from an input longitude.
        SECTION 2.15: Generate LPS Grid values 
        SECTION 2.16: Generate LTM Grids values 
        SECTION 2.17: Generate LGRS Grid values 
        SECTION 2.18: Generate LGRS polar grid values  
    SECTION 3: MAIN PROGRAM
        SECTION 3.1: determine if input arguments are correct
        SECTION 3.2: Read in input arguments center x center y or corners,
        SECTION 3.3: Converting Input Coordinates to Common Grid Format.
        SECTION 3.4: Generate the Correct Grid to the desired resolution
        SECTION 3.5: Convert grid Coordinates to correct output format.
        SECTION 3.6: Determine grid cell names in specified output format
        SECTION 3.7: Assign output WKT1 Projection string for writing file
        SECTION 3.8:  Write grid to ESRI shape file and format output names

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
# License X/MIT
Copyright (C) <date> <USGS, Astrogeology Science Center>

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the 
"Software"), to deal in the Software without restriction, including 
without limitation the rights to use, copy, modify, merge, publish, 
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
 in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of USGS shall not be used
in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization from USGS.'''

# =====================================================================
# SECTION 0: INITIALIZE PROGRAM AND CALL EXTERNAL LIBRARIES


import os                   # used as system utilities
import sys
                                            # os -> Python 3.11.4
                                            # sys 3.11.4

import numpy as np          # To perform mathematical calculations
                            # and uses numpy arrays for all grid 
                            # positions
                                            # np 1.25.2

import osgeo                # used to support geospatial data 
                            # processing and formatting. 

from osgeo import gdal,ogr,osr  # lib for rasters, vectors, and SRS
                                # Coordinate systems
                                # osgeo -> '3.7.1'
osr.DontUseExceptions() # suppress GDAL versioning warnings

from LGRS_Coordinate_Conversion_mk6 import *       
                                # LGRS conversion lib for 
                                # converting between 
                                # LatLon, LTM, LPS, LGRS, LGRS_ACC
                    # LGRS_Coordinate_Conversion_mk3 -> mk3

import re                   # for input coordinate parsing
                            # re -> '2.2.1'

import datetime             # To perform time monitoring
                                            # datetime -> Python 3.11.4

ProgName="LGRS_Grid_Generation_mk6"

# =====================================================================
# SECTION 1: REFERENCE SURFACE PARAMETERS
# Geodetic function constants for grid and grid generation on the Moon.
# He we pre-assign the correct projection values and pre-dimension 
# rectangular grids.

# ---------------------------------------------------------------------
# SECTION 1.1: LIST COORDINATE FORMATS


def initialize_LGRS_grid_generation_globals():
    """
    Code creates globalized values to be called in the main program.
    Globals placed here, so constants can be easily changed in the 
    future.

    Globals Defined:
    (list: str) coordinate_format: List of all the coordinate formats that
                                are able to be read into this program.
    (str) coordinate_format_str: List of input coordinate formats for 
                                printing.
    (list: str) grid_format: List of all grid and grid resolutions that
                            this program is capable of generating.
    (str) grid_format_str: List of grid and grid resolutions in a string
                           for printing 
    (re: str) LTM_pattern: re pattern used for reading in LTM coordinates 
    (re: str) LTM_pattern_full: re pattern used for reading in LTM coordinates 
                    used at the grid corners that are condensed together
    (re: str) LGRS_pattern: re pattern used for reading condensed LGRS grid  
                    coordinates used at the grid corners. 
    (re: str) PolarLGRS_pattern: re pattern used for reading condensed LGRS 
                    grid coordinates used at the grid corners on the 
                    lunar poles.
    (re: str) LGRS_ACC_pattern: re pattern used for reading LGRS 
                        Artemis Condensed Coordinates (ACC) without  
                        truncation. Coordinates are located at the
                        grid corners.
    (re: str) PolarLGRS_ACC_pattern: re pattern used for reading LGRS 
                        Artemis Condensed Coordinates (ACC) without  
                        truncation. Coordinates are located at the
                        grid corners on the lunar pulls
    (re: str) LPS_pattern_full: re pattern used for reading in LPS 
                            coordinates
    """

    global coordinate_format, coordinate_format_str, grid_format
    global grid_format_str, LTM_pattern, LTM_pattern_full, LGRS_pattern
    global PolarLGRS_pattern, LGRS_ACC_pattern, PolarLGRS_ACC_pattern
    global LPS_pattern_full

    # input formats 
    coordinate_format = ["LatLon","PolarLatLon","LTM","LPS","LGRS",
                       "PolarLGRS","LGRS_ACC","PolarLGRS_ACC"]
    coordinate_format_str = "LatLon\nPolarLatLon\nLTM\nLPS\nLGRS\n\
                           PolarLGRS\nLGRS_ACC\nPolarLGRS_ACC\n\
                           ACC and PolarACC can be used for \n\
                           output Coordinates only."

    # LGRS and 25km_all grid
    grid_format = ["border","global","global_all","25km","25km_all","1km","100m","10m","1m"]
    grid_format_str = "border\nglobal\nglobal_all\n25km\n25km_all\n1km\n100m\n10m\n1m"

    # supports floating or truncated integer values
    LTM_pattern = r'(\d+)([NS]{1})'
    LTM_pattern_full = r'(\d+)([NS]{1})(\d+)(E)(\d+)(N)'

    LPS_pattern_full = r'([NS]{1})(\d+(?:\.\d+)?|\.\d+)(E)(\d+(?:\.\d+)?|\.\d+)(N)'


    # supports only truncated values (this is ok; non-compressed 
    # Coordinates can be read in as a full floating point)
    LGRS_pattern = r'(\d+)([A-Z]{1})([A-Z]{1})([A-Z]{1})(\d{5})(\d{5})'
    PolarLGRS_pattern = r'([A-Z]{1})([A-Z]{1})([A-Z]{1})(\d{5})(\d{5})'
    LGRS_ACC_pattern = r'(\d+)([A-Z]{1})([A-Z]{1})([A-Z]{1})([-A-Z]{1})(\d+)([-A-Z]{1})(\d+)'
    PolarLGRS_ACC_pattern = r'([A-Z]{1})([A-Z]{1})([A-Z]{1})([-A-Z]{1})(\d+)([-A-Z]{1})(\d+)'


# ---------------------------------------------------------------------
# SECTION 1.2: DEFINE MAP PROJECTIONS FOR PCRS AND GRID SYSTEMS USING
#              WKT1 STINGS
    """
    Define the the output map projections as WKT1 stings. Used when
    Exporting the files so they are georeferenced and display correctly
    on the lunar surface 

    Globals Defined:
    (WKT1: str) N_LPS_projection_meters LPS northern zone, Defined by 
                    stereographic projection,
                    projection center at 90°, 0°,
                    Scale factor of 0.994
                    False origins of 500km E, 500km N, and 
                    units in meters.
    (WKT1: str) S_LPS_projection_meters LPS Southern zone, Defined by 
                    stereographic projection,
                    projection center at -90°, 0°,
                    Scale factor of 0.994
                    False origins of 500km E, 500km N, and 
                    units in meters.
    (WKT1: str) N_LTM_projection_meters LTM northern zones, Defined by
                    transverse Mercator projection, 
                    projection center at 0° lam0 (longitude),
                    Scale factor of 0.999
                    False origins of 250km E, 2500km N, and 
                    units in meters. lam0 is unique in each zone.
    (WKT1: str) S_LTM_projection_meters LTM southern zones, Defined by 
                    transverse Mercator projection,
                    projection center at 0° lam0,
                    Scale factor of 0.999
                    False origins of 250km E, 2500km N, and 
                    units in meters. lam0 is unique in each zone.
    (WKT1: str) Display_geocentric_degrees formats projection
                    for the lunar spheroid with a semi major axis of 
                    1737400m with no associated flattening, f=0.
                    Units are output in arc degrees.
                    Projection may very based on default projection 
                    in GIS software.
    """
    global N_LPS_projection_meters,S_LPS_projection_meters
    global N_LTM_projection_meters,S_LTM_projection_meters
    global Display_geocentric_degrees

    # METERS <- LTM, LPS, LGRS, and ACC LGRS
    N_LPS_projection_meters = '''PROJCS["Moon (2015) - Sphere / Ocentric / North Polar",
    GEOGCS["Moon (2015) - Sphere / Ocentric",DATUM["Moon (2015) - Sphere",SPHEROID["Moon (2015) - Sphere",1737400,0,
    AUTHORITY["IAU","30100"]],AUTHORITY["IAU","30100"]],PRIMEM["Reference Meridian",0,AUTHORITY["IAU","30100"]],
    UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["IAU","30100"]],
    PROJECTION["Polar_Stereographic"],
    PARAMETER["latitude_of_origin",90],
    PARAMETER["central_meridian",0],
    PARAMETER["scale_factor",0.994],
    PARAMETER["false_easting",500000],
    PARAMETER["false_northing",500000],
    UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],
    AUTHORITY["ARTEMIS","30197"]]'''

    S_LPS_projection_meters = '''PROJCS["Moon (2015) - Sphere / Ocentric / South Polar",
    GEOGCS["Moon (2015) - Sphere / Ocentric",DATUM["Moon (2015) - Sphere",SPHEROID["Moon (2015) - Sphere",1737400,0,
    AUTHORITY["IAU","30100"]],AUTHORITY["IAU","30100"]],PRIMEM["Reference Meridian",0,AUTHORITY["IAU","30100"]],
    UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["IAU","30100"]],
    PROJECTION["Polar_Stereographic"],
    PARAMETER["latitude_of_origin",-90],
    PARAMETER["central_meridian",0],
    PARAMETER["scale_factor",0.994],
    PARAMETER["false_easting",500000],
    PARAMETER["false_northing",500000],
    UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],
    AUTHORITY["ARTEMIS","30198"]]'''

    N_LTM_projection_meters  =  '''PROJCS["Moon (2015) - Sphere / Ocentric / Tranverse Mercator",
    GEOGCS["Moon (2015) - Sphere / Ocentric",DATUM["Moon (2015) - Sphere",SPHEROID["Moon (2015) - Sphere",1737400,0,
    AUTHORITY["IAU","30100"]],AUTHORITY["IAU","30100"]],PRIMEM["Reference Meridian",0,AUTHORITY["IAU","30100"]],
    UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["IAU","30100"]],
    PROJECTION["Transverse_Mercator"],
    PARAMETER["latitude_of_origin",0],
    PARAMETER["central_meridian",{}],
    PARAMETER["scale_factor",0.999],
    PARAMETER["false_easting",250000],
    PARAMETER["false_northing",0],
    UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],
    AUTHORITY["ARTEMIS","30195"]]'''

    S_LTM_projection_meters = '''PROJCS["Moon (2015) - Sphere / Ocentric / Tranverse Mercator",
    GEOGCS["Moon (2015) - Sphere / Ocentric",DATUM["Moon (2015) - Sphere",SPHEROID["Moon (2015) - Sphere",1737400,0,
    AUTHORITY["IAU","30100"]],AUTHORITY["IAU","30100"]],PRIMEM["Reference Meridian",0,AUTHORITY["IAU","30100"]],
    UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["IAU","30100"]],
    PROJECTION["Transverse_Mercator"],
    PARAMETER["latitude_of_origin",0],
    PARAMETER["central_meridian",{}],
    PARAMETER["scale_factor",0.999],
    PARAMETER["false_easting",250000],
    PARAMETER["false_northing",2500000],
    UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],
    AUTHORITY["ARTEMIS","30196"]]'''

    # DEGREES <- planetocentric latitude and longitude
    Display_geocentric_degrees = '''GEOGCS["Moon (2015) - Sphere / Ocentric",DATUM["Moon (2015) - Sphere",
    SPHEROID["Moon (2015) - Sphere",1737400,0,AUTHORITY["IAU","30100"]],AUTHORITY["IAU","30100"]],
    PRIMEM["Reference Meridian",0,AUTHORITY["IAU","30100"]],UNIT["degree",0.0174532925199433,
    AUTHORITY["EPSG","9122"]],AXIS["Latitude",NORTH],AXIS["Longitude",EAST],AUTHORITY["IAU","30100"]]'''

# ---------------------------------------------------------------------  
# SECTION 1.3: Define the LTM LPS coordinate limits and resolution
    """
    Define the map projection and lunar spheroid constants to be used
    in dimensioning the grids. 

    Globals Defined:
    (float) ResX grid plotting resolution in arc-degrees for X dimension
    (float) ResY grid plotting resolution in arc-degrees for Y dimension
    (float, int) LTM_S_Boarder LTM southern projection boundary 
    (float, int) LTM_Equator planetocentric latitude
    (float, int) LTM_N_Boarder LTM northern projection boundary 
    (float, int) LPS_S_Boarder most northern latitude for the S pole projection area
    (float, int) LPS_N_Boarder most southern latitude for the N pole projection area
    (float, int) LatLon_W_Meridian planetocentric min longitude
    (float, int) LatLon_E_Meridian planetocentric max longitude
    (float, int) LUNAR_RADIUS Lunar reference spheroid radial value
    (float, int) LatLon_Central_Meridian Prime meridian of the moon
    (float, int) LPS_S_phi0 LPS map projection center latitude value for S pole
    (float, int) LPS_N_phi0 LPS map projection center latitude value for N pole
    (float, int) LTM_Width LTM zone width of 8°
    (float, int) FalseEasting LTM False Easting (Shift in X)
    (float, int) FalseNorthing LTM False Northing (Shift in Y)
    (float, int) FalseEasting_polar LPS False Easting (Shift in X)
    (float, int) FalseNorthing_polar LPS False Northing (Shift in Y)
    """

    global ResX,ResY
    global LTM_S_Boarder,LTM_Equator,LTM_N_Boarder,LPS_S_Boarder
    global LPS_N_Boarder,LatLon_W_Meridian,LatLon_E_Meridian,LUNAR_RADIUS
    global LatLon_Central_Meridian,LPS_S_phi0,LPS_N_phi0,LTM_Width
    global FalseEasting,FalseNorthing,FalseEasting_polar,FalseNorthing_polar
    
    # Define the shape file point sampling frequency in arc-degrees.
    # these arguments are not called for grids projected in rectangular
    # Coordinates eg: LPS or LTM as the grid lines are vertical and 
    # horizontal in the right projection. This is used for plots that
    # cover multiple zones
    ResX = 0.01
    ResY = 0.01
 
    # define the LTM zone boarders  
    LTM_S_Boarder = -82
    LTM_Equator = 0
    LTM_N_Boarder = 82
    LTM_Width = 8
    FalseEasting = 250E3
    FalseNorthing = 2500E3

    # Lat Long boarders
    LatLon_W_Meridian = -180
    LatLon_Central_Meridian = 0
    LatLon_E_Meridian = 180
    LUNAR_RADIUS = 1737400

    #define the LPS zone boarders
    LPS_S_phi0 = -90
    LPS_S_Boarder = -80
    LPS_N_phi0 = 90
    LPS_N_Boarder = 80
    FalseEasting_polar = 500E3
    FalseNorthing_polar = 500E3


# =====================================================================
# SECTION 2: FUNCTION LIBARY

# SECTION 2.1: Naming attribute table when creating shape files
# SECTION 2.2: Assign values to attribute table
# SECTION 2.3: Export multi polygon grid to esri shape-file
# SECTION 2.4: Export single polygon, as a multi polygon data type, to
# SECTION 2.5: Assign X Y Coordinates to multi polygon osgeo ogr
# SECTION 2.6: Assign X Y positional Coordinates to single ogr ring
# SECTION 2.7: Assign X Y positional Coordinates to ogr rings in multi
#              polygon geometry
# SECTION 2.8: Intersect ogr multi polygon geometries to a single geometry
# SECTION 2.9: Merge ogr multi polygon geometries to a single geometry
# SECTION 2.10: Clip multi polygon geometries within another geometry
# SECTION 2.11: Extract X, Y positions from multi polygon geometry
# SECTION 2.12: Calculate hemisphere location on a sphere in degrees
# SECTION 2.13: calculate LTM zone prime meridian from a LTM zone number
# SECTION 2.14: Determine the LTM zone number from an input longitude.
# SECTION 2.15: Generate LPS Grid values 
# SECTION 2.16: Generate LTM Grids values 
# SECTION 2.17: Generate LGRS Grid values 
# SECTION 2.18: Generate LGRS polar grid values  
    
# ---------------------------------------------------------------------
# SECTION 2.1: Naming attribute table when creating shape files


def initialize_attribute_table(layer, attribute=True):
    """
    Initialize the attribute table of an esri shapefile.

    Input parameters:
    (str) Layer - osgeo osr layer definition
    (bool) attribute - switch to initialize the shapefile attribute table

    Defined Parameters:
    (int) id - polygon count - unitless
    (str) LTM | LPS | LGRS | PolarLGRS | LGRS_ACC | PolarLGRS_ACC | ACC | 
    PolarACC - Grid coordinate deigned at center of grid area
    (int) centX - center of grid square in Eastings -unit: deg or m
    (int) centY - center of grid square in Northings -unit: deg or m
    (int) ULX - upper left corner X -unit: arc-degrees or meters
    (int) ULY - upper left corner Y -unit: arc-degrees or meters
    (int) LRX - unit: lower right corner X  designed-arc-degrees or meters
    (int) LRY - unit: lower right corner Y -arc-degrees or meters
    More can be added if needed 

    Returns:
    (str) Layer - osgeo osr layer definition with attributes 
                    initialized 
    """
    if attribute:
        # polygon count
        field_label = ogr.FieldDefn('id', ogr.OFTInteger)
        layer.CreateField(field_label)

        # defined grid coordinate
        if form_out == "LTM":
            field_label = ogr.FieldDefn('LTM', ogr.OFTString)
            layer.CreateField(field_label)

        elif form_out == "LPS":
            field_label = ogr.FieldDefn('LPS', ogr.OFTString)
            layer.CreateField(field_label)

        elif ( ( form_out == "LGRS" ) or \
              ( form_out == "PolarLGRS" ) ):
            field_label = ogr.FieldDefn('LGRS', ogr.OFTString)
            layer.CreateField(field_label) 

        elif ( ( form_out == "LGRS_ACC" ) or \
               ( form_out == "PolarLGRS_ACC" ) ):
            field_label = ogr.FieldDefn('LGRS_ACC', ogr.OFTString)
            layer.CreateField(field_label) 

        elif ( ( form_out == "ACC" ) or \
               ( form_out == "PolarACC" ) ):
            
            field_label = ogr.FieldDefn('ACC', ogr.OFTString)
            layer.CreateField(field_label) 

        else:
            raise ValueError ("Coordinate could not be assigned"
                              "attribute table.")
        
        # center Coordinates of the grid area
        field_label = ogr.FieldDefn('centX', ogr.OFTInteger)
        layer.CreateField(field_label)
        
        field_label = ogr.FieldDefn('centY', ogr.OFTInteger)
        layer.CreateField(field_label)

        # define corner points of the grid area
        field_label = ogr.FieldDefn('ULX', ogr.OFTInteger)
        layer.CreateField(field_label)
        
        field_label = ogr.FieldDefn('ULY', ogr.OFTInteger)
        layer.CreateField(field_label)

        field_label = ogr.FieldDefn('LRX', ogr.OFTInteger)
        layer.CreateField(field_label)
            
        field_label = ogr.FieldDefn('LRY', ogr.OFTInteger)
        layer.CreateField(field_label)
    
    return layer

# ---------------------------------------------------------------------
# SECTION 2.2: Assign values to attribute table


def set_field_values(feat, k, coord, X, Y, attribute):
    """
    Initialize the attribute table of an esri shapefile.

    Input perimeters:
    (str) feat - osgeo osr polygon feature, part of a layer
    (int) k - index referenced to a value assigned to a specific polygon 
              in the total polygon count
    (bytes) coord - Grid area name, truncated to specified value
    (array; doubles) X - Grid X or easting values -units:m
    (array; doubles) Y - Grid Y or northing values -units:m
    (bool) attribute - switch to initialize the shapefile attribute table

    Assigned Parameters:
    (int) id - polygon count - unitless
    (str) LTM | LPS | LGRS | PolarLGRS | LGRS_ACC | PolarLGRS_ACC | ACC | 
    PolarACC - Grid coordinate deigned at center of grid area
    (int) centX - center of grid square in Eastings -unit: deg or m
    (int) centY - center of grid square in Northings -unit: deg or m
    (int) ULX - upper left corner X -unit: arc-degrees or meters
    (int) ULY - upper left corner Y -unit: arc-degrees or meters
    (int) LRX - unit: lower right corner X  -arc-degrees or meters
    (int) LRY - unit: lower right corner Y -arc-degrees or meters
    More can be added if needed 

    Returns:
    (str) Layer - osgeo osr layer definition with attributes 
                    initialized 
    """
    if attribute:
        # set unique number for each polygon
        feat.SetField('id', k)

        # assign coordinate value for the output format to table
        if form_out == "LTM":
            feat.SetField('LTM', coord)

        elif form_out == "LPS":
            feat.SetField('LPS', coord)

        elif ( ( form_out == "LGRS" ) or \
              ( form_out == "PolarLGRS" ) ):
            feat.SetField('LGRS', coord)

        elif ( ( form_out == "LGRS_ACC" ) or \
               ( form_out == "PolarLGRS_ACC" ) ):
            feat.SetField('LGRS_ACC', coord)

        elif ( ( form_out == "ACC" ) or \
               ( form_out == "PolarACC" ) ):
            feat.SetField('ACC', coord)

        else:
            raise ValueError ("Coordonate could not be assigned"
                              "attribute table.")

        # calculate grid area extrema
        n = np.nanmax(Y)
        s = np.nanmin(Y)
        e = np.nanmax(X)
        w = np.nanmin(X)
        
        # calculate center coordinate value
        CentX = (w + e)/2
        CentY = (s + n)/2

        # assign center coordinate value to table
        feat.SetField('centX', round(CentX))
        feat.SetField('centY', round(CentY))

        # assign grid corner coordinates to table
        feat.SetField('ULX', round(w))
        feat.SetField('ULY', round(n))
        feat.SetField('LRX', round(e))
        feat.SetField('LRY', round(s))

        return feat

# ---------------------------------------------------------------------  
# SECTION 2.3: Export multi polygon grid to esri shapefile


def Xport_multipolygons(X, Y, GridName, f_o,
                        WKT1,
                        attribute=True,
                        output_file_form="shp",
                        overwrite=False):
    """
    This section of code reads X Y coordinates of degrees of meters and
    writes them to an esri shapefile.This code writes out a multipoly-
    gone geometry, by writing multiple osgeo ogr rings. it cannot out
    put a single polygon layer. Writing of the attribute table is
    optional. Additionally, the projection is assigned using a WKT1
    string, and can be changed to a more custom projection if needed.
    Map projection WKT1 stings are assigned at the beginning of this
    program.

    Input Perimeters:
    (array; doubles) X - Grid X or easting values -units:m
    (array; doubles) Y - Grid Y or northing values -units:m
    (bytes) GridName - Grid area name, truncated to specified value
    (str) f_o - output file name and output folder
    (str) WKT1 - Well-Known Text 1 formatted string with output projection
    (bool) attribute - switch to initialize the shapefile attribute table
    (str) output_file_form - specify file output format
                           - sph for esri shapefile
                           - gpkg for geopackage
    (bool) overwrite - switch to initialize that a shapefile be 
                       overwritten if it already exists

    Returns:
    (file) - esri shapefiles (shp,shx,dbf,prj) with name of f_o
           - gpgk and other files are also witten if a diferent 
           - file format is specified 
    (folder) - esri shapefiles containing folder with name of f_o
    """
    # Procedure to remove a shape file if it exists to its 
    # file name can be reused when rewriting a new shapefile

    if output_file_form == "shp":
        if overwrite:
            if os.path.isfile(f_o + "/" + f_o + ".shp"):
                print("file:",f_o + "/" + f_o + ".shp","exists removing and rewriting")
                os.remove(f_o + "/" + f_o + ".shp")
                os.remove(f_o + "/" + f_o + ".shx")
                os.remove(f_o + "/" + f_o + ".dbf")
                os.remove(f_o + "/" + f_o + ".prj")
        else:
            # If overwriting is not specified, then the program terminates 
            if os.path.isfile(f_o + "/" + f_o + ".shp"):
                print("file:",f_o + "/" + f_o + ".shp","exists terminating.")
                raise ValueError ("Terminating File cannot be overwritten. ")
                exit()

        # initialize shape file for exporting
        driver = ogr.GetDriverByName('ESRI Shapefile')
        ds  =  driver.CreateDataSource(f_o)
        srs = osr.SpatialReference()
        srs.ImportFromWkt(WKT1)
        layer = ds.CreateLayer(f_o, srs, geom_type=ogr.wkbMultiPolygon)

    # format for a geopackage file type
    elif output_file_form == "gpkg":
        if overwrite:
            if os.path.isfile(f_o + ".gpkg"):
                print("file:",f_o + ".gpkg","exists removing and rewriting")
                os.remove(f_o + ".gpkg")

        else:
            # If overwriting is not specified, then the program terminates 
            if os.path.isfile(f_o + ".gpkg"):
                print("file:", f_o + ".gpkg","exists terminating.")
                raise ValueError ("Terminating File cannot be overwritten. ")
                exit()
                
        driver=ogr.GetDriverByName('GPKG')
        ds = driver.CreateDataSource(f_o+".gpkg")
        srs=osr.SpatialReference()
        srs.ImportFromWkt(WKT1) # assigning projection to file
        layer = ds.CreateLayer(f_o,srs,geom_type=ogr.wkbMultiPolygon)
    
    # Catch all if the output format could not be determined
    else:
        raise ValueError("Cannot write output file: Output format could "
                         "not be determined")

    # Geopackage as example if you want to support an alternative format
    #driver=ogr.GetDriverByName('GPKG')
    #ds = driver.CreateDataSource(f_o+".gpkg")
    #srs=osr.SpatialReference()
    #srs.ImportFromWkt(WKT1) # assigning projection to file
    #layer = ds.CreateLayer(f_o,srs,geom_type=ogr.wkbMultiPolygon)

    # set up attributes layer from external function
    layer = initialize_attribute_table(layer,attribute)

    # get layer feature definition
    feature_defn = layer.GetLayerDefn()

    # loop through grid and assigning grid areas to geometry
    k = 0
    for i in range(0, X.shape[0]):   # change line
        PolygonString = ogr.Geometry(ogr.wkbPolygon)
        ring = ogr.Geometry(ogr.wkbLinearRing)

        if np.isnan(X[i,0]): # skip if empty ring
            continue

        for j in range(0, X.shape[1]): # iterate though line lines
            if not np.isnan(X[i,j]): # if data value exits continue
                                     # nan is used to end the grid 
                ring.AddPoint(X[i,j], Y[i,j])
        ring.CloseRings()

        # add geometry to sting
        PolygonString.AddGeometry(ring)

        # add polygon to multi polygon
        MultiPolygonString = ogr.Geometry(ogr.wkbMultiPolygon)
        MultiPolygonString.AddGeometry(PolygonString)

        # assign attributes to shapefile if specified
        feat = ogr.Feature(feature_defn)
        if attribute:
            # function to assign attribute values using 
            # an external function
            feat=set_field_values(feat,
                             k,
                             GridName[i,0].decode(),
                             X[i,:],
                             Y[i,:],
                             attribute)
        
        # assign polygons to feature
        feat.SetGeometry(MultiPolygonString)
        layer.CreateFeature(feat)
    
        # iterate id value for next shape file
        k += 1

        # clean up memory and continue for next grid area to be 
        # dimensioned
        feat = None
    ds = layer = feat = None

# ---------------------------------------------------------------------
# SECTION 2.4: Export single polygon, as a multi polygon data type, to
#              an ESRI shapefile. 

def Xport_polygon(X, Y, GridName, f_o,
                  WKT1,
                  attribute=True,
                  output_file_form="shp",
                  overwrite=False):
    """
    This section of code reads X Y coordinates of degrees of meters and
    writes them to an  shapefile or geopack.This code writes out a multipoly-
    gone geometry for a single polygon ogr ring. Writing of the attribute 
    table is optional. Additionally, the projection is assigned using 
    a WKT1 string, and can be changed to a more custom projection if
    needed. Map projection WKT1 stings are assigned at the beginning of
    this program.

    Input Perimeters:
    (array; doubles) X - Grid X or easting values -units:m
    (array; doubles) Y - Grid Y or northing values -units:m
    (bytes) GridName - Grid area name, truncated to specified value
    (str) f_o - output file name and output folder
    (str) WKT1 - WKT1 formatted string with output projection
    (bool) attribute - switch to initialize the shape file attribute table
    (str) output_file_form - specify file output format
                           - sph for esri shapefile
                           - gpkg for geopackage
    (bool) overwrite - switch to initialize that a shapefile be 
                       overwritten if it already exists

    Returns:
    (file) - esri shapefiles (shp,shx,dbf,prj) with name of f_o
           - gpgk and other files are also written if a different 
           - file format is specified 
    (folder) - esri shapefiles containing folder with name of f_o
    """
    
    # Procedure to remove a shape file if it exists to its 
    # file name can be reused when rewriting a new shapefile
    if output_file_form == "shp":
        if overwrite:
            if os.path.isfile(f_o + "/" + f_o + ".shp"):
                print("file:",f_o + "/" + f_o + ".shp","exists removing and rewriting")
                os.remove(f_o + "/" + f_o + ".shp")
                os.remove(f_o + "/" + f_o + ".shx")
                os.remove(f_o + "/" + f_o + ".dbf")
                os.remove(f_o + "/" + f_o + ".prj")
        else:
            # If overwriting is not specified, then the program terminates 
            if os.path.isfile(f_o + "/" + f_o + ".shp"):
                print("file:",f_o + "/" + f_o + ".shp","exists terminating.")
                raise ValueError ("Terminating File cannot be overwritten. ")
                exit()

        # initialize shape file for exporting
        driver = ogr.GetDriverByName('ESRI Shapefile')
        ds  =  driver.CreateDataSource(f_o)
        srs = osr.SpatialReference()
        srs.ImportFromWkt(WKT1)
        layer = ds.CreateLayer(f_o, srs, geom_type=ogr.wkbMultiPolygon)

    # format for a geopackage file type
    elif output_file_form == "gpkg":
        if overwrite:
            if os.path.isfile(f_o + ".gpkg"):
                print("file:",f_o + ".gpkg","exists removing and rewriting")
                os.remove(f_o + ".gpkg")

        else:
            # If overwriting is not specified, then the program terminates 
            if os.path.isfile(f_o + "/" + f_o + ".gpkg"):
                print("file:",f_o + "/" + f_o + ".gpkg","exists terminating.")
                raise ValueError ("Terminating File cannot be overwritten. ")
                exit()
                
        driver=ogr.GetDriverByName('GPKG')
        ds = driver.CreateDataSource(f_o+".gpkg")
        srs=osr.SpatialReference()
        srs.ImportFromWkt(WKT1) # assigning projection to file
        layer = ds.CreateLayer(f_o,srs,geom_type=ogr.wkbMultiPolygon)
    
    # Catch all if the output format could not be determined
    else:
        raise ValueError("Cannot write output file: Output format could "
                         "not be determined")

    # set up attributes layer 
    layer = initialize_attribute_table(layer,attribute)

    # get layer feature definition
    feature_defn = layer.GetLayerDefn()

    # Create polygon
    PolygonString = ogr.Geometry(ogr.wkbPolygon)
    ring = ogr.Geometry(ogr.wkbLinearRing)

    # assign attributes to shapefile if specified
    for i in range(0, X.shape[0]):   # change grid area
        for j in range(0, X.shape[1]):   # iterate polygon area
            ring.AddPoint(X[i,j], Y[i,j])
    ring.CloseRings()

    # add geometry to polygon 
    PolygonString.AddGeometry(ring)

    # assign polygons to feature
    MultiPolygonString = ogr.Geometry(ogr.wkbMultiPolygon)
    MultiPolygonString.AddGeometry(PolygonString)

    k = 1 # assign id value
    
    # assign values to attribute table if specified 
    feat = ogr.Feature(feature_defn)
    if attribute:
        # function to assign attribute values using 
        # an external function
        feat = set_field_values(feat,
                                k,
                                GridName[0,0].decode(),
                                X[0,:],
                                Y[0,:],
                                attribute)
    # assign polygons to feature
    feat.SetGeometry(MultiPolygonString)
    layer.CreateFeature(feat)

    # clean up memory and end function
    ds = layer = feat = None

# ---------------------------------------------------------------------
# SECTION 2.5: Assign X Y Coordinates to multi polygon odgeo ogr 
#              geometry. 


def assign_xy_grid_to_multipolygon(X, Y):
    """
    This section of code reads X Y coordinates of degrees of meters and
    writes them to an osgeo ogr multiplication geometry. Map projection
    does not matter, however, please not, any clipping or additional 
    operations will need to be completed in the exact dame coordinate
    format to avoid issues.This call only works with grid formatted 
    areas, thus a bounding box area will not work

    Input Perimeters:
    (2D array; doubles) X - Grid X or easting values -units:m
    (2D array; doubles) Y - Grid Y or northing values -units:m

    Returns:
    (class) - MultiPolygonString multi-poly string of a grid area
    """
    # create new multi-polygon geometry
    MultiPolygonString = ogr.Geometry(ogr.wkbMultiPolygon)

    # Assign X and Y values to geometry, 1 in removed from the total 
    # looping size as this is plotted like a grid, and all four corner
    # points are assigned at the same time
    for i in range(0,X.shape[0]-1): # iterate through grid areas
        for j in range(0,X.shape[1]-1): # iterate through grid bounding area

            # create geometry to store location information
            PolygonString = ogr.Geometry(ogr.wkbPolygon)
            ring = ogr.Geometry(ogr.wkbLinearRing)
            
            # add all four corner points of the grid area to ring
            ring.AddPoint(X[i,j], Y[i,j])
            ring.AddPoint(X[i,j+1], Y[i,j+1])
            ring.AddPoint(X[i+1,j+1], Y[i+1,j+1])
            ring.AddPoint(X[i+1,j], Y[i+1,j])
            ring.AddPoint(X[i,j], Y[i,j])
            ring.CloseRings() # close ring

            # add geometry to single polygon
            PolygonString.AddGeometry(ring)

            # add polygon to multi polygon
            MultiPolygonString.AddGeometry(PolygonString)

    # export multi polygon
    return MultiPolygonString

# ---------------------------------------------------------------------
# SECTION 2.6: Assign X Y positional Coordinates to single ogr ring
#              geometry

def assign_xy_to_polygon (X, Y):
    """
    This section of code reads X Y coordinators of degrees of meters and
    writes them to an osgeo ogr polygon geometry. Map projection
    does not matter, however, please not, any clipping or additional 
    operations will need to be completed in the exact dame coordinate
    format to avoid issues.This call only works with a single polygon,
    however, the polygon geometry can be longer that the 5 values LGRS
    grids are written in. For grids see the 
    assign_xy_grid_to_multipolygon function.

    Input Perimeters:
    (2D array; doubles) X - Grid X or easting values -units:m
    (2D array; doubles) Y - Grid Y or northing values -units:m

    Returns:
    (class) - PolygonString single polygon geometry string of a grid 
              area
    """
    # create new polygon geometry and ring to assign location values to
    PolygonString = ogr.Geometry(ogr.wkbPolygon)
    RingString = ogr.Geometry(ogr.wkbLinearRing)

    # Assign X and Y values to ring
    for i in range(0, X.shape[0]): # loop through polygons
        for j in range(0, X.shape[1]): # loop through geometry 
            if np.isnan(X[i,j]):
                continue
            RingString.AddPoint(X[i,j], Y[i,j]) # assign value
            
    # finalize polygon geometry
    RingString.CloseRings()

    # add ring geometry to polygon
    PolygonString.AddGeometry(RingString)

    # export finalized polygon
    return PolygonString

# ---------------------------------------------------------------------
# SECTION 2.7: Assign X Y positional Coordinates to ogr rings in multi
#              polygon geometry

def assign_arbitrary_xy_to_multipolygon (X, Y):
    """
    This section of code reads X Y coordinators of degrees of meters and
    writes them to an osgeo ogr multipolygon geometry. Map projection
    does not matter, however, please not, any clipping or additional 
    operations will need to be completed in the exact dame coordinate
    format to avoid issues.For grids see the 
    assign_xy_grid_to_multipolygon function. For single polygons see
    the assign_xy_to_polygon function.

    Input Perimeters:
    (2D array; doubles) X - Grid X or easting values -units:m
    (2D array; doubles) Y - Grid Y or northing values -units:m

    Returns:
    (class) - PolygonString single polygon geometry string of a grid 
              area
    """
    # create new polygon geometry and ring to assign location values to
    # Create a new multipolygon geometry
    MultiPolygon = ogr.Geometry(ogr.wkbMultiPolygon)
    
    # Iterate through each polygon
    for i in range(X.shape[0]):  # Loop through polygons
        if np.isnan(X[i, 0]):
            continue
        # Create a new polygon geometry and ring
        PolygonString = ogr.Geometry(ogr.wkbPolygon)
        RingString = ogr.Geometry(ogr.wkbLinearRing)
        
        for j in range(X.shape[1]):  # Loop through points in the polygon
            if np.isnan(X[i, j]):
                continue
            RingString.AddPoint(X[i, j], Y[i, j])  # Add point to ring
        
        # Close the ring
        RingString.CloseRings()
        
        # Add ring geometry to polygon
        PolygonString.AddGeometry(RingString)
        
        # Add the polygon to the multipolygon
        MultiPolygon.AddGeometry(PolygonString)
    
    return MultiPolygon

# ---------------------------------------------------------------------
# SECTION 2.8: Intercect ogr multi polygon geometries to a single geometry

def intersect_multipolygons(multipolygon1, multipolygon2):
    """Intersect two MULTIPOLYGON Z geometries and return the result."""
    # Create an empty multipolygon to store the result
    intersect_multipolygon = ogr.Geometry(ogr.wkbMultiPolygon)
    
    # Iterate over each polygon in the first multipolygon
    for i in range(multipolygon1.GetGeometryCount()):
        polygon1 = multipolygon1.GetGeometryRef(i)
        # Iterate over each polygon in the second multipolygon
        for j in range(multipolygon2.GetGeometryCount()):
            polygon2 = multipolygon2.GetGeometryRef(j)
            
            # Perform intersection between the two polygons
            intersected_polygon = polygon1.Intersection(polygon2)
            # Add the intersected polygon to the result if it's not empty
            if not intersected_polygon.IsEmpty():
                intersect_multipolygon.AddGeometry(intersected_polygon)
    
    return intersect_multipolygon

# ---------------------------------------------------------------------
# SECTION 2.9: Merge ogr multi polygon geometries to a single geometry

def merge_multipolygons(geometries1, geometries2):
    """
    The merge_multipolygons combined the two geometries of polygon or
    multi polygon format into a single multi polygon format. The function
    function recursively merges multi polygons within each polygon 
    geometry, so every polygon is re assigned to the a new geometry.
    It checks if the geometry is a multi polygon, and if so, recursively 
    constructs a new polygons. This process continues until all levels of 
    multi polygons are merged into a single ogr geometry object 
    representing the merged multi polygon. Map projection or the Coordinates
    does not matter, however, please not, any clipping or additional 
    operations will need to be completed in the exact dame coordinate
    format to avoid issues.

    Input Perimeters:
    (class) geometries1 - set of geometry values formatted as an ogr multi or
                        single polygon
    (class) geometries2 - set of geometry values formatted as an ogr multi or
                        single polygon

    Returns:
    (class) merged_multipolygon - multi polygon composed of both geometries1
                                  and geometries2.
    """
    # create new multipolygon geometry
    merged_multipolygon = ogr.Geometry(ogr.wkbMultiPolygon)
    
    # Iterate over geometries from the first set
    for geometry in geometries1:
        # Check if the geometry is a multipolygon
        if geometry.GetGeometryType() == ogr.wkbMultiPolygon:
            # If it's a multipolygon, recursively merge its constituent polygons
            merged_multipolygon.AddGeometry(merge_multipolygons(geometry, []))
        else:
            # If it's a polygon, add it directly to the merged multipolygon
            merged_multipolygon.AddGeometry(geometry)
    
    # Iterate over geometries from the second set
    for geometry in geometries2:
        # Check if the geometry is a multipolygon
        if geometry.GetGeometryType() == ogr.wkbMultiPolygon:
            # If it's a multipolygon, recursively merge its constituent polygons
            merged_multipolygon.AddGeometry(merge_multipolygons(geometry, []))
        else:
            # If it's a polygon, add it directly to the merged multipolygon
            merged_multipolygon.AddGeometry(geometry)
    
    # return new multipolygon of merger areas
    return merged_multipolygon

# ---------------------------------------------------------------------
# SECTION 2.10: Clip multipolygon geometries within another geometry


def clip_multipolygon_by_polygon(multipolygon_geometry, clip_polygon):
    """
    The clip_multipolygon_by_polygon takes two geometries, the 
    multipolygon_geometry (ogr multipolygon) and the clip_polygon 
    (ogr polygon) in order to clip the multipolygon_geometry to be
    contained within the clip_geometry. This function has the capability
    of removing polygon areas, or cutting them to ensure that if a 
    polygon has a portion of its area in the clip geometry is retained
    in the final 

    Input Perimeters:
    (class) multipolygon_geometry - set of geometry values formatted as 
                        an ogr multi or single polygon
    (class) clip_polygon -set of geometry values formatted as an ogr
                        multi or single polygon

    Returns:
    (class) merged_multipolygon - multipolygon composed of
                    multipolygon_geometry within the region of
                    clip_polygon.
    """
    # Create a new empty geometry to store the clipped result
    clipped_geometry = ogr.Geometry(ogr.wkbMultiPolygon)

    # Iterate over each polygon in the multipolygon geometry
    for i in range(multipolygon_geometry.GetGeometryCount()):
        polygon = multipolygon_geometry.GetGeometryRef(i)

        # Perform intersection with the clip polygon
        clipped_polygon = polygon.Intersection(clip_polygon)

        # Add the clipped polygon to the result if it's not empty
        if not clipped_polygon.IsEmpty():
            clipped_geometry.AddGeometry(clipped_polygon)

    return clipped_geometry

# ---------------------------------------------------------------------
# SECTION 2.11: Extract X, Y potions from multipolygon geometry 


def extract_xy_from_multipolygon(multipolygon_geometry):
    """
    extract_xy_from_multipolygon takes a multipolygon input geometry,
    and extracts the X and Y position and returns them as a numpy array.
    
    Input Perimeters:
    (class) multipolygon_geometry - set of geometry values formatted as an ogr multi or
                        single polygon

    Returns:
    (class) merged_multipolygon - multipolygon composed of
                    multipolygon_geometry within the region of
                    clip_polygon.
    """
    # Initialize lists to store X and Y coordinates
          
    # calculate the max number of points in a polygon, this is used
    # to dimension the arrays and to save memory
    max_points = 0
    for i in range(multipolygon_geometry.GetGeometryCount()):
        polygon = multipolygon_geometry.GetGeometryRef(i)
        ring = polygon.GetGeometryRef(0)
        points2 = ring.GetPointCount()
        if points2 > max_points:
            max_points = points2

    # Dimension arrays by number of polygons and max number of points  
    # in polygon
    x_values = np.full((multipolygon_geometry.GetGeometryCount(), max_points), np.nan)
    y_values = np.full((multipolygon_geometry.GetGeometryCount(), max_points), np.nan)

    # Iterate over each polygon in the multipolygon and 
    # assign geometry X, Y to arrays
    k = 1  # keep track of total polygon count
    for i in range(multipolygon_geometry.GetGeometryCount()):
        polygon = multipolygon_geometry.GetGeometryRef(i)
        ring = polygon.GetGeometryRef(0)
        points = ring.GetPointCount()
        
        # Extract X and Y coordinates from the polygon and assign
        # to numpy array
        for j in range(0,points):
            x, y, _ = ring.GetPoint(j)
            x_values[i,j] = x
            y_values[i,j] = y

        k+=1 # iterate k 
    
    return x_values, y_values

# ---------------------------------------------------------------------
# SECTION 2.12: Calculate hemisphere location on a sphere in degrees
#               Return numeric value

def Calc_hemisphere(phi):
    """
    determine the correct lunar hemisphere given a position in latitude
    
    Input Perimeters:
    (float, int) phi - planetocentric latitude in the range of 
                       -90° - 90°.

    Returns:
    (str) hemisphere - return string value of the hemisphere location
                         either "N" for northern or "S" for southern.
    """
    if phi >= 0: # We assume the equator to be northern
        return "N"
    elif phi < 0:
        return "S"
    else:
        raise ValueError("Hemisphere could not be determined") 

# ---------------------------------------------------------------------
# SECTION 2.13: calculate LTM zone prime meridian from a LTM zone number


def Calc_LTM_prime_meridian(zone, type):
    """
    Determine the correct LTM central meridian planetocentric longitude
    given an LTM number in the range of [1-45]. Central meridian can be
    returned in either radian of arc-degrees
    
    Input Perimeters:
    (int) zone - LTM 8° W zone number. Northern and Southern zones and 
                 their difference not utilized.
    (str) typ - the type of output, radians or degrees, needed from
                this calculation

    Returns:
    (float) lam0 - longitude of LTM zone central meridian.
    """
    # safety checks
    if zone > 45:
        raise ValueError ("LTM zone 46\{N|S\} of higher does not exist")
    elif zone < 0:
        raise ValueError ("Negative LTM zones do not exist")

    # determine the type of calculation required
    if type == "rad":
        lam0 = np.deg2rad((zone-1) * (LTM_Width) - 180. + (LTM_Width/2))    
    elif type == "deg":
        lam0 = (zone-1) * (LTM_Width) - 180. + (LTM_Width/2)
    else:
        raise ValueError ("LTM zone prime meridian could not be "
                          "calculated: place specify type "
                          "\{\"rad\"|\"deg\"\}")

    return lam0

# ---------------------------------------------------------------------
# SECTION 2.14: Determine the LTM zone number from an input longitude.


def Calc_LTM_zone(X):
    """
    Determine the correct LTM zone number from an input longitude.

    Input Perimeters:
    (int, float) X - LTM formatted in plenetocentric -180 to 180 degrees.
                Northern and Southern zones and their difference not
                utilized.

    Returns:
    (int) zone - LTM zone number without hemisphere designation.
    """
    zone = np.floor((np.mean(X) + 180)/(LTM_Width)) + 1
    return int(zone)

# ---------------------------------------------------------------------
# SECTION 2.15: Generate LPS Grid values according to a specified 
#              resolution and coordinate type


def generate_LPS_grid(Grid_Res, h):
    """
    This section of code creates all Lunar Polar Stereographic (LPS)
    systems grid given a small number of inputs. 

    Input Perimeters:
    (str) Grid_Res - Grid format and resolution
    (str) h - hemisphere (either north, "N" | south, "S")

    Valid types of grids
    Grid_Res == border - Single polygon of the LPS 80° boarder plotted
                        plotted in Latitude and Longitude coordinates.
                        This grid is for display in a global context 
                        as box is drawn around the plot area.
    Grid_Res == global - Single polygon of the LPS 80° boarder plotted
                        in LPS Coordinates. Data is exported in the 
                        LPS projections depending on the zone 
                        location. This plot can only be plotted in 
                        polar stereographic. No points are placed
                        at the projection centers (90° or-90°) and 
                        will display as a line if in the wrong 
                        projection.
    Returns:
    (2D array, floats) GridX - Longitude or Easting values of the LPS 
                               boarder.
    (2D array, floats) GridY - Latitude or Northing values of the LPS
                               boarder.
    """
    # plot LPS areas for display in LatLon. This plots a box 
    # in latitude and longitude values.
    if Grid_Res == "border":
        # Southern zone
        if h == "S":
            # grid-lines W->E
            GridX_T = np.arange(LatLon_W_Meridian,LatLon_E_Meridian,ResX)[:,np.newaxis] 
            GridX_T = np.append(GridX_T,LatLon_E_Meridian)
            
            GridY_T = np.ones(GridX_T.shape[0]-1).T*LPS_S_Boarder
            GridY_T = np.append(GridY_T,LPS_S_Boarder) #S 

            # draw E->W values
            GridX_B = np.arange(LatLon_E_Meridian,LatLon_W_Meridian,-ResX)[:,np.newaxis]
            GridX_B = np.append(GridX_B,LatLon_W_Meridian)
            
            GridY_B = np.ones(GridX_B.shape[0]-1).T*LPS_S_phi0
            GridY_B = np.append(GridY_B,LPS_S_phi0) #S 
            
            # Compile values into a list
            X_vals = (GridX_T,GridX_B)
            Y_vals = (GridY_T,GridY_B)

        # If in the northern zone
        elif h == "N":
            # gridlines W->E
            GridX_T = np.arange(LatLon_W_Meridian,LatLon_E_Meridian,ResX)[:,np.newaxis] 
            GridX_T = np.append(GridX_T,LatLon_E_Meridian)
            
            GridY_T = np.ones(GridX_T.shape[0]-1).T*LPS_N_Boarder
            GridY_T = np.append(GridY_T,LPS_N_Boarder)

            # draw E->W values
            GridX_B = np.arange(LatLon_E_Meridian,LatLon_W_Meridian,-ResX)[:,np.newaxis] 
            GridX_B = np.append(GridX_B,LatLon_W_Meridian)
            
            GridY_B = np.ones(GridX_B.shape[0]-1).T*LPS_N_phi0
            GridY_B = np.append(GridY_B,LPS_N_phi0)#S 

            # combine grid values into a list 
            X_vals = (GridX_T,GridX_B)
            Y_vals = (GridY_T,GridY_B)

        # If no zone could be identified
        else:
            raise ValueError("LPS display Generation Terminated, Hemisphere could "
                            "not be identified.")
        
        # Concatenate grid values into a single list of values
        GridX = np.concatenate(X_vals)
        GridY = np.concatenate(Y_vals)

    # generate grid that surrounds the LPS 80° boarder 
    # plotted in LPS Coordinates. This plots a circle in
    # rectangular Coordinates
    elif Grid_Res == "global":
        if h == "S":
            GridX = np.arange(LatLon_W_Meridian,LatLon_E_Meridian,ResX)[:,np.newaxis] #grid lines
            GridX = np.append(GridX,LatLon_E_Meridian)

            GridY = np.ones(GridX.shape[0]-1).T*LPS_S_Boarder
            GridY = np.append(GridY,LPS_S_Boarder)
        
        elif h == "N":
            GridX = np.arange(LatLon_W_Meridian,LatLon_E_Meridian,ResX)[:,np.newaxis] #grid lines
            GridX = np.append(GridX,LatLon_E_Meridian)

            GridY = np.ones(GridX.shape[0]-1).T*LPS_S_Boarder
            GridY = np.append(GridY,LPS_S_Boarder)

        else:
            raise ValueError("LPS Generation Terminated, Hemisphere and "
                            "grid type could not be identified.")
    # For program functionality and for exporting.
    # Output vectors are transformed from a 1D row vector 
    # into a 2D array with a single row of values.
    GridX = GridX[np.newaxis,:]
    GridY = GridY[np.newaxis,:]

    # output LPS grids
    return GridX, GridY

# ---------------------------------------------------------------------
# SECTION 2.16: Generate LTM Grids values according to a specified 
#              resolution and coordinate type


def generate_LTM_grid(Grid_Res,z=None,
                               h=None):
    """
    This function generates grid for the Lunar Transverse Mercator (LTM)
    system. 

    Input Perimeters:
    (int, float) Grid_Res - Grid format and resolution
    (int) z - LTM zone number designation
    (str) h - LTM hemispheric location designation

    Types of grids
    Grid_Res == global_all - Generate a multipolygon grid of all 90 LTM 
                            zones plotted and dimensioned in Lat Lon 
                            coordinate format with their associated. Grid
                            area is plotted as a box, but should display 
                            correctly the Lat Lon, LTM and LPS 
                            projections.  
    Grid_Res == global - Generate a single polygon of single LTM zone 
                        plotted in Lat Lon in the associated LTM 
                        projection. This boarder is drawn as a box and 
                        can be displaced in LatLon, LPS and LTM 
                        projections.
    Grid_Res == border - Generate a single polygon of single LTM zone 
                        plotted in LTM Coordinates in the associated LTM 
                        projection. This boarder is drawn as a box and 
                        can be displaced in LatLon, LPS and LTM 
                        projections but is designed for LTM only.
        Note: Global and Boarder grids are both written in latitude
              and longitude, the coordinate values are converted later
              in the function code to proper coordinate format.

    Returns:
    (2D array, floats) GridX -  Longitude or Easting values of the LTM
                               boarder.
    (2D array, floats) GridY -  Latitude or Northing values of the LTM
                               boarder.
    """
    # generate grid for all LTM zones in LatLon
    if Grid_Res == "global_all":
        k = 0 # unique id number for each polygon processed in 
            # this function

        # Calculate dimensioned of the array
        # Top and bottom, 8° width subdivided by ResX covered twice
        # Left and right 82° range subdivided by ResY per zone covered twice
        # Add for for last appended ending values
        # each row is a polygon
        # each column is an LTM zone, 90 in total
        ArrayX = int(4 + (2*LTM_Width)/ResX + (2*82)/ResY) # dimension all points in
                                                     # grid area
        ArrayY = 90 # dimension for total number of polygons

        # dimension numpy arrays for storing and exporting grid values.
        GridX = np.zeros((ArrayY,ArrayX))
        GridY = np.zeros((ArrayY,ArrayX))

        # Loop through LTM grid areas based on LTM dimensions and generate
        # grid values for plotting

        # Iterate through the latitude range
        for i in range(LTM_S_Boarder,LTM_N_Boarder,LTM_N_Boarder):
            # iterate through the longitude range
            for j in range(LatLon_W_Meridian,LatLon_E_Meridian,LTM_Width):
                
                # determine LTM zone boundaries
                n = i + LTM_N_Boarder
                s = i
                e = j + LTM_Width
                w = j

                #write out grid values in W->E (Bottom)
                GridX_B = np.arange(w,e,ResX)[:,np.newaxis] 
                GridX_B = np.append(GridX_B,w)
                
                GridY_B = np.ones(GridX_B.shape[0]-1).T*s
                GridY_B = np.append(GridY_B,s)

                #write out grid values in S -> N (right)
                GridY_R = np.arange(s,n,ResY)[:,np.newaxis]
                GridY_R = np.append(GridY_R,n)
                
                GridX_R = np.ones(GridY_R.shape[0]-1).T*e
                GridX_R = np.append(GridX_R,e)

                #write out grid values in E -> W (Top)
                GridX_T = np.arange(e,w,-ResX)[:,np.newaxis]
                GridX_T = np.append(GridX_T,w)
                
                GridY_T = np.ones(GridX_T.shape[0]-1).T*n
                GridY_T = np.append(GridY_T,n)

                #write out grid values in N -> S (left)
                GridY_L = np.arange(n,s,-ResY)[:,np.newaxis]
                GridY_L = np.append(GridY_L,s)
                
                GridX_L = np.ones(GridY_L.shape[0]-1).T*w
                GridX_L = np.append(GridX_L,w)

                # combine values into list
                X_vals = (GridX_B,GridX_R,GridX_T,GridX_L)
                Y_vals = (GridY_B,GridY_R,GridY_T,GridY_L)

                # concatenate values into single list 
                # for exporting 
                GridX[k,:] = np.concatenate(X_vals)
                GridY[k,:] = np.concatenate(Y_vals)

                k+=1 # iterate to the next polygon

    # generate grid for a single LTM zone in LatLon, 
    # converted later to proper exporting format.
    elif Grid_Res == "border" or Grid_Res == "global":
        # Calculate dimensions of the array:
        # Top and bottom, 8° width subdivided by ResX covered twice
        # Left and right 82° range subdivided by ResY per zone covered twice
        # Add for for last appended ending values
        # each row is a polygon
        # each column is an LTM zone, 1 in total
        ArrayX = int(4 + (2*LTM_Width)/ResX + (2*82)/ResY) 

        # Generate arrays based on calculated size
        GridX = np.zeros((ArrayX))
        GridY = np.zeros((ArrayX))

        # determine zone bounds in the north south range
        if h == "N": # if in the northern hemisphere of zones, phi >=0°
            n = LTM_N_Boarder
            s = LTM_Equator
        elif h  ==  "S": # if in the northern hemisphere of zones, phi <0°
            n = LTM_Equator - 1E-13   # integers calculated in adjacent zone.
                                  # this is done correctly. 
                                  # To plot correctly we "force the value"
                                  # into the correct zone bu making is a 
                                  # very small value at matching precision
            s =  LTM_S_Boarder
        else:
            # if hemisphere could not be determined
            raise ValueError ("Correct LTM zone dimensions could not "
                              "be determined.")
        
        # calculate the east boarder value of the LTM zone
        e =  Calc_LTM_prime_meridian(z,"deg") + LTM_Width/2 - 1E-13
            # similarly, we subtract a small value to endure the
            # east zone is plotted correctly
        
        # calculate the west boarder value for the LTM zone
        w = Calc_LTM_prime_meridian(z,"deg") - LTM_Width/2

        # Write out grid X and Y values

        # bottom L -> R of the LTM zone
        GridX_B = np.arange(w,e,ResX)
        GridY_B = np.ones(GridX_B.shape[0])*s

        GridX_B = np.append(GridX_B,e)
        GridY_B = np.append(GridY_B,s)

        # right side of the LTM zone
        GridY_R = np.arange(s,n,ResY)
        GridX_R = np.ones(GridY_R.shape[0])*e

        GridX_R = np.append(GridX_R,e)
        GridY_R = np.append(GridY_R,n)

        # top edge of LTM zone
        GridX_T = np.arange(e,w,-ResX)
        GridY_T = np.ones(GridX_T.shape[0])*n

        GridX_T = np.append(GridX_T,w)
        GridY_T = np.append(GridY_T,n)

        #left side of the LTM zone
        GridY_L = np.arange(n,s,-ResY)
        GridX_L = np.ones(GridY_L.shape[0])*w

        GridY_L = np.append(GridY_L,s)
        GridX_L = np.append(GridX_L,w)

        # format into single row of vales
        X_vals = (GridX_B,GridX_R,GridX_T,GridX_L)
        Y_vals = (GridY_B,GridY_R,GridY_T,GridY_L)

        # concatenate values into a single list
        GridX[:] = np.concatenate(X_vals)
        GridY[:] = np.concatenate(Y_vals)

        # Convert Grid from 1D row vector into a 2D array 
        # with a single row of values. Extra array axis
        # added to support exporting in this program 
        GridX =  GridX[np.newaxis,:]
        GridY =  GridY[np.newaxis,:]

    else:
        raise ValueError("LTM Generation Terminated, Grid type could "
                         "not be determined.")
    # return LTM grid area
    return GridX, GridY

# ---------------------------------------------------------------------
# SECTION 2.17: Generate LGRS Grid values according to a specified 
#              resolution and coordinate type


def generate_LGRS_grid(Grid_Res,lonBand=None,
                                h=None,
                                z1=None,
                                h1=None,
                                ulx=None,
                                uly=None,
                                z2=None,
                                h2=None,
                                lrx=None,
                                lry=None,
                                multigrid=False,
                                LatLon_override=False):
    """
    This function is used to generate associated grids for the Lunar Grid
    Reference System (LGRS) in all resolutions and formats.
    
    Input Perimeters:
    (int, float) Grid_Res - Grid format and resolution to be generated
    (int) lonBand -  LTM zone number designator for LGRS plot for grids
              25km or greater in resolution.
    (str) h - LTM hemispheric location designator for LGRS plot for grids
              25km or greater in resolution.
    (int) z1 - LTM zone number designator for upper left hand coordinate
               of grid dimensioning bounding box. Only used in grid less-
               than or equal to 1km
    (str) h1 - LTM hemisphere designator for upper left hand coordinate
               of grid dimensioning bounding box. Only used in grid less-
               than or equal to 1km
    (double) ulx - upper left x coordinate of the grid dimensioning
                   bounding box. Only used in grids less than or equal
                   to 1km.
    (double) uly - upper left y coordinate of the grid dimensioning
                   bounding box. Only used in grids less than or equal
                   to 1km.
    (int) z2 - LTM zone number designator for lower right hand coordinate
               of grid dimensioning bounding box. Only used in grid less-
               than or equal to 1km
    (str) h2 - LTM hemisphere designator for lower right hand coordinate
               of grid dimensioning bounding box. Only used in grid less-
               than or equal to 1km
    (double) lrx - Lower right x coordinate of the grid dimensioning
                   bounding box. Only used in grids less than or equal
                   to 1km.
    (double) lry - Lower right y coordinate of the grid dimensioning
                   bounding box. Only used in grids less than or equal
                   to 1km.
    (bool) multigrid - logical value used to create a grid that crosses
                       multiple zones in the LTM portion of LGRS only. 
                       Once activated and set to True, the output format
                       will be overwritten and forced to a LatLon 
                       coordinate structure in the defined projection for 
                       latitude. The coordinates will be forced to lat
                       lon in this setting regardless if the grid is in
                       only one zone. If False, z1, z2, h1, and h2 will
                       be check to ensure the grid is only in one zone.
                       Program will terminate if the grid is in multi-
                       pile zones.
    (bool) LatLon_override - Logical value that will force the inverse 
                             conversion of grid position back to lunar
                             latitude and longitude. the final exported 
                             grid will be be referenced to to lat lon.
                             NOTE: the potential for incorrect projections,
                             the LGRS and LGRS grids are meant to be 
                             utilized in the LPS and LTM PCRS. Without
                             utilization map grid distances will very and
                             map distortion will be calibrated. It is 
                             recommended that this option not be utilized 
                             to ensure correct positioning with the 
                             designed LGRS PCRS (LPS and LTM).
                             
    Types of grids
    Grid_Res == global - Generate multipolygon grid of the global area
                         grid of the LGRS plot within a single specified
                         LTM zone. Typically this will generate 1, 8°x
                         10°zone (labeled C or X) and 9, 8°x8° zones and 
                         is plotted in LTM Coordinates in a LatLon projection.
                         multi grid not supported as it is not needed.
    Grid_Res == global_all - Generate multipolygon grid of all global area
                         grids in LGRS. This plots across all LTM zones in
                         the latitude longitude coordinate format and in
                         the associated projection. This setting results 
                         in the generation of 90 8°x10° LGRS global areas,
                         and 810 8°x8° zones.
                         multi grid not supported as it is not needed.
    Grid_Res == 25km_all - Generate multipolygon grid of all 25km grid
                         zones of LGRS. This plots is applied to the either
                         LTM portion of LGRS and is calculated in the
                         latitude longitude coordinate format with 
                         the associated projection. This setting results 
                         in the generation of 69920 25km grid zones.
                         multigrid not supported as it is not needed.
    Grid_Res == 25km - Generate multipolygon grid of all 25km grid
                     zones in a single LTM zone of the LGRS. This
                     grid is calculated in the LTM Coordinates in with 
                     the associated projection.
                     multigrid not supported as it is not needed.
    Grid_Res == 1km - Generate multipolygon 1km  LGRS grid in 
                     a single or across multiple LTM zones. Grid is 
                     created in LTM Coordinates in with the 
                     associated projection. Upper left and lower right 
                     corner point required, multigrid is supported.
    Grid_Res == 100m - Generate multipolygon 100m  LGRS grid in 
                     a single or across multiple LTM zones. Grid is 
                     created in LTM Coordinates in with the 
                     associated projection. Upper left and lower right 
                     corner point required, multigrid is supported.
    Grid_Res == 10m - Generate multipolygon 10m  LGRS grid in 
                     a single or across multiple LTM zones. Grid is 
                     created in LTM Coordinates in with the 
                     associate projection. Upper left and lower right 
                     corner point required, multigrid is supported.
    Grid_Res == 1m - Generate multipolygon 1m  LGRS grid in 
                     a single or across multiple LTM zones. Grid is 
                     created in LTM Coordinates in with the 
                     associated projection. Upper left and lower right 
                     corner point required, multigrid is supported.
        Note: This function is used to calculate the output grids for
              LGRS associated coordinate formats including LGRS, 
              LGRS_ACC ( a full Artemis condensed format), and ACC
              with global location truncation. 

    Returns:
    (2D array, floats) GridX - Longitude or Easting values of the LGRS 
                               grid system contained in LTM coordinates 
                               and contained within the correct LTM
                               boarder.
    (2D array, floats) GridY - Latitude or Northing values of the LGRS 
                               grid system contained in LTM coordinates 
                               and contained within the correct LTM
                               boarder.
    (bool) multigrid - if user specified processing a multigrid shape
                       file return state to main function for later
                       processing.
    """

    # generate LGRS global area grid in LTM Coordinates
    if Grid_Res == "global": # LGRS 8° Areas in LTM Coordinates
                           # Coordinates
        # Calculate dimensions of the array
        # Top and bottom, 10° width subdivided by ResX covered twice (oversized)
        # Left and right 8° range subdivided by ResY per zone covered twice
        # Add for for last appended ending values
        # each row is a polygon
        # each column is an LTM zone, 1 in total
        ArrayX = int(4+(2*LTM_Width)/ResX+(2*(LTM_Width+2))/ResY)
        ArrayY = int(10)

        # generate grids arrays to assign and export values from
        # NaN values are used to signify the grid polygon is complete
        # or that there is not a polygon in the row. This format is 
        # used to support polygon clipping and merging
        GridX = np.full((ArrayY,ArrayX),np.nan)
        GridY = np.full((ArrayY,ArrayX),np.nan)

        # Determine LGRS  bounds based on LTM projection areas
        if h == "N": # if in the northern portion
            N = LTM_N_Boarder-2 # provides a 0° to 80° range. Comparability for
                              # 10° zones
            S = LTM_Equator
        elif h == "S": # if in the southern portion
            N = LTM_Equator
            S = LTM_S_Boarder+2 # provides a 0° to 80° range. Compatibility for
                              # 10° zones
        else:
            # Raise problem is hemisphere could not be determined
            raise ValueError ("Correct LTM zone dimensions for LGRS grid "
                              "be determined.")
        
        k = 0 # unique identifier for polygon values
        # iterate through the north south range of the LGRS global 
        # area grid, ie: 8° in latitude
        for i in range(S,int(N),LTM_Width):

            # determine the central meridian of the LTM projection 
            # area for processing
            lam0 = Calc_LTM_prime_meridian(lonBand,"deg")

            # determine LGRS global area bounds
            n = i + LTM_Width # top
            s = i # bottom
            e = lam0 + LTM_Width/2 - 1E-13 # right, forced into zone by s
                                     # subtracting small value
            w = lam0 - LTM_Width/2  # left

            # add in extended range, 8°x10°, LGRS global areas for
            # the north south range
            if s == -80:
                s =- 82
            elif n == 80:
                n = 82
            elif n == 0 and h == "S":
                n -= 1E-13

            # generate arrays values for grid

            # write our L -> R to B
            GridX_B = np.arange(w,e,ResX)
            GridY_B = np.ones(GridX_B.shape[0])*s

            GridX_B = np.append(GridX_B,e)
            GridY_B = np.append(GridY_B,s)

            # write our B -> T to R
            GridY_R = np.arange(s,n,ResY)
            GridX_R = np.ones(GridY_R.shape[0])*e

            GridX_R = np.append(GridX_R,e)
            GridY_R = np.append(GridY_R,n)

            # write our R -> L to T
            GridX_T = np.arange(e,w,-ResX)
            GridY_T = np.ones(GridX_T.shape[0])*n

            GridX_T = np.append(GridX_T,w)
            GridY_T = np.append(GridY_T,n)            
            
            # write our T -> B to L
            GridY_L = np.arange(n,s,-ResY)
            GridX_L = np.ones(GridY_L.shape[0])*w

            GridX_L = np.append(GridX_L,w)
            GridY_L = np.append(GridY_L,s)

            # combine all grid values into a list
            X_vals = (GridX_B,GridX_R,GridX_T,GridX_L)
            Y_vals = (GridY_B,GridY_R,GridY_T,GridY_L)

            # concatenate positioning values into a single list
            # this is stack where each row is a single polygon
            X_vals = np.concatenate(X_vals)
            Y_vals = np.concatenate(Y_vals)

            # determine output array bounds
            X_vals_len = X_vals.shape[0]
            Y_vals_len = Y_vals.shape[0]

            # based on bounds assign value to polygon grid
            GridX[k,0:X_vals_len] = X_vals[0:X_vals_len]
            GridY[k,0:Y_vals_len] = Y_vals[0:Y_vals_len]

            k+=1 # iterate polygon identifier number

    # 8°x8° and 8°x10° LGRS Areas in LatLon Coordinates 
    elif Grid_Res == "global_all": 

        # Calculate dimensions of the array
        # Top and bottom, 10° width subdivided by ResX covered twice (oversized)
        # Left and right 8° range subdivided by ResY per zone covered twice
        # Add for for last appended ending values
        # each row is a polygon
        # All grid areas in Y. 360°/8°zones*20 grid areas per zone=900
        ArrayX = int(4 + (2*LTM_Width)/ResX + (2*(LTM_Width + 2))/ResY)
        ArrayY = 360*LTM_Width*20 

        # generate grids arrays to assign and export values from
        # NaN values are used to signify the grid polygon is complete
        # or that there is not a polygon in the row. This format is 
        # used to support polygon clipping and merging
        GridX = np.full((ArrayY,ArrayX),np.nan)
        GridY = np.full((ArrayY,ArrayX),np.nan)

        k = 0 # assign value to iterate through polygon values
        for i in range(int(LTM_S_Boarder) + 2,int(LTM_N_Boarder) - 2,LTM_Width):
            for j in range(int(LatLon_W_Meridian),int(LatLon_E_Meridian),LTM_Width): 
                
                n = i + LTM_Width
                s = i
                e = j + LTM_Width - 1E-13
                w = j

                # add in extended range zones
                if s == -80:
                    s =- 82
                elif n == 80:
                    n = 82
                elif n == 0 and h == "S":
                    n -= 1E-13

                # L -> R B
                GridX_B = np.arange(w,e,ResX)
                GridY_B = np.ones(GridX_B.shape[0])*s

                GridX_B = np.append(GridX_B,e)
                GridY_B = np.append(GridY_B,s)

                # B -> T R
                GridY_R = np.arange(s,n,ResY)
                GridX_R = np.ones(GridY_R.shape[0])*e

                GridX_R = np.append(GridX_R,e)
                GridY_R = np.append(GridY_R,n)

                # R -> L T
                GridX_T = np.arange(e,w,-ResX)
                GridY_T = np.ones(GridX_T.shape[0])*n

                GridX_T = np.append(GridX_T,w)
                GridY_T = np.append(GridY_T,n)            
                
                # T -> B L
                GridY_L = np.arange(n,s,-ResY)
                GridX_L = np.ones(GridY_L.shape[0])*w

                GridX_L = np.append(GridX_L,w)
                GridY_L = np.append(GridY_L,s)

                # combine parts of the polygon into a list
                X_vals = (GridX_B,GridX_R,GridX_T,GridX_L)
                Y_vals = (GridY_B,GridY_R,GridY_T,GridY_L)

                # concatenate polygons together into a single 
                # row
                X_vals = np.concatenate(X_vals)
                Y_vals = np.concatenate(Y_vals)

                # determine number of grid values for assignment
                X_vals_len = X_vals.shape[0]
                Y_vals_len = Y_vals.shape[0]

                # assign grid values to array of polygons
                GridX[k,0:X_vals_len] = X_vals[0:X_vals_len]
                GridY[k,0:Y_vals_len] = Y_vals[0:Y_vals_len]

                k+=1 # iterate to the next polygon 

    # grid of 25km grid zone in all LTM zones in LatLon
    elif Grid_Res == "25km_all": 

        # Calculate dimensions of the array
        # Top and bottom, 10° width subdivided by ResX covered twice (oversized)
        # Left and right 8° range subdivided by ResY per zone covered twice
        # Add for for last appended ending values
        # each row is a polygon
        # All grid areas in Y. 360°/8°zones*20 grid areas per zone=900
                #   + 100 for extra global areas
        # 90 LTM zones as well

        ArrayX = int(10/ResX/2+100) # 10 degrees divided by resolution. 
                            # a safe number to represent the curvature of an
                            # LGRS LTM zone.
        
        ArrayY = int(FalseEasting/25E3 * FalseNorthing/25E3 - 240 + 100) 
                        # No 25km grid squares in LTM zone
                        # the max is 1000 per LTM zone
                        # but most are clipped and 760
                        # should be the max
        
        ArrayZ = int(360/LTM_Width*2)  # 90 LTM zones

        # dimension arrays,( LTM zones 
        #                  x 25km grids zones per LTM zone
        #                  x estimated max polygon length )
        # NaNs are used to close the polygon or used to show
        # now polygon was assigned
        GridX = np.full((ArrayZ,ArrayY,ArrayX),np.nan) 
        GridY = np.full((ArrayZ,ArrayY,ArrayX),np.nan) 

        k = 0 # for iterating through unique polygons

        # Loop through all LTM zones 
        for i in range(int(LTM_S_Boarder),int(LTM_N_Boarder),int(LTM_N_Boarder)):
            for j in range(int(LatLon_W_Meridian),int(LatLon_E_Meridian),LTM_Width): 

                z = Calc_LTM_zone(j + LTM_Width/2)
                lam0 = Calc_LTM_prime_meridian(z,"deg")

                n = i + LTM_N_Boarder
                s = i
                e = j + LTM_Width - 1E-13
                w = j

                if s < 0:
                    h = 'S'
                elif s >= 0:
                    h = 'N'

                if n == 0 and h == "S":
                    n -= 1E-13

                # dimensions used for dimensioning data array
                X_max = FalseEasting
                X_min = 0
                Y_max = FalseNorthing
                Y_min = 0

                # generate arrays
                Grid_X = np.arange(X_min,X_max,25E3)
                Grid_Y = np.arange(Y_min,Y_max,25E3)
                
                Grid_X = np.append(Grid_X,X_max)
                Grid_Y = np.append(Grid_Y,Y_max)

                # format meshgrid for all zones
                Cylindrical_X,Cylindrical_Y = np.meshgrid(Grid_X,Grid_Y)

                # shift X coordinates so grid is centered in LTM zone
                Cylindrical_X+=FalseEasting/2

                # assign grid to geometry
                LGRS_25km_geom_single_area = assign_xy_grid_to_multipolygon(Cylindrical_X,Cylindrical_Y)

                # collect global areas and merge 
                GlobalX,GlobalY,multigrid = generate_LGRS_grid("global",z,h)
                # convert global zone boundary coordinates from latlon to LTM 
                for m in range(0,GlobalX.shape[0]):
                    for n in range(0,GlobalY.shape[1]):
                        if np.isnan(GlobalX[m,n]):
                            continue
                        z,h,GlobalX[m,n],GlobalY[m,n] = toLTM(GlobalX[m,n],GlobalY[m,n],zone=None,trunc_val=0,process_errors=False)
                # assign to mutlipolygon
                LGRS_Global_geom = assign_arbitrary_xy_to_multipolygon(GlobalX,GlobalY)
                # union areas into 
                LGRS_25kmGlobal_geom_single_area = intersect_multipolygons(LGRS_25km_geom_single_area,LGRS_Global_geom)

                # determine LTM bounding box for clipping
                LTM_X,LTM_Y = generate_LTM_grid("border",z,h)

                # convert coordinators to LTM values
                for m in range(0,LTM_X.shape[0]):
                    for n in range(0,LTM_X.shape[1]):
                        z,h,LTM_X[m,n],LTM_Y[m,n] = toLTM(LTM_X[m,n],LTM_Y[m,n],zone=None,trunc_val=0,process_errors=False)                    
                LTM_zone_clip_geom = assign_xy_to_polygon(LTM_X,LTM_Y)

                # clip to bounding box
                LGRS_area_clipped = clip_multipolygon_by_polygon(LGRS_25kmGlobal_geom_single_area,LTM_zone_clip_geom)

                # extract X, Y positions from clipped region
                X_vals,Y_vals = extract_xy_from_multipolygon(LGRS_area_clipped)

                GridX[k,0:X_vals.shape[0],0:X_vals.shape[1]] = X_vals
                GridY[k,0:X_vals.shape[0],0:X_vals.shape[1]] = Y_vals

                # convert coordinators to Lat Lon values for exporting
                for mi in range(0,GridX.shape[1]):
                    for ni in range(0,GridX.shape[2]):
                        GridX[k,mi,ni],GridY[k,mi,ni] = toLatLon(GridX[k,mi,ni],GridY[k,mi,ni],z,h,process_errors=True)

                k+=1

        # Reshape the 3D array to a 2D array
        GridX = np.vstack(GridX) #[0,:,:] #.reshape(GridX.shape[0], -1)
        GridY = np.vstack(GridY) #[0,:,:] #.reshape(GridY.shape[0], -1)
  
    # calculate all 25km grid areas in a single LGRS LTM zone in LTM 
    # coordinates
    elif Grid_Res == "25km": 
        if lonBand == None or h == None:
            raise ValueError ("LGRS 25km single global zone is projected "
                              "A LTM zone is required for plotting "
                              "operation: Z\{1-45\} h\{N|S\}")

        # Calculate dimensions of the array
        # Top and bottom, 10° width subdivided by ResX covered twice (oversized)
        # Left and right 8° range subdivided by ResY per zone covered twice
        # Add for for last appended ending values
        # each row is a polygon
        # All grid areas in Y. 360°/8°zones*20 grid areas per zone=900
        # 90 LTM zones as well
        ArrayX = int(10/ResX/2 +18010) # 10 degrees divided by resolution. 
                            # a safe number to represent the curvature of an
                            # LGRS LTM zone.
        
        ArrayY = int(FalseEasting/25E3*FalseNorthing/25E3 - 240 + 1000) 
                        # No 25km grid squares in LTM zone
                        # the max is 1000 per LTM zone
                        # but most are clipped and 760
                        # should be the max

        # dimension grid values
        GridX = np.full((ArrayY,ArrayX),np.nan) 
        GridY = np.full((ArrayY,ArrayX),np.nan) 

        # determine LTM zone boundaries based on hemisphere
        if h == "N":
            n = LTM_N_Boarder
            s = LTM_Equator
        elif h == "S":
            n = LTM_Equator - 1E-13   # integers calculated in adjacent zone.
                                  # this is done correctly. 
                                  # To plot correctly we "force the value"
                                  # into the correct zone
            s = LTM_S_Boarder
        else:
            raise ValueError ("Correct LTM zone dimensions could not "
                              "be determined.")
        
        # determine east west range base on the LGRS lonBand
        e = Calc_LTM_prime_meridian(lonBand,"deg")+LTM_Width/2-1E-13
        w = Calc_LTM_prime_meridian(lonBand,"deg")-LTM_Width/2

        # Generate 25km grid values
        X_max = FalseEasting
        X_min = 0
        Y_max = FalseNorthing
        Y_min = 0

        # generate arrays
        Grid_X = np.arange(X_min,X_max,25E3)
        Grid_Y = np.arange(Y_min,Y_max,25E3)
        
        Grid_X = np.append(Grid_X,X_max)
        Grid_Y = np.append(Grid_Y,Y_max)

        # format mesh grid for all zones
        Cylindrical_X,Cylindrical_Y = np.meshgrid(Grid_X,Grid_Y)
        Cylindrical_X += FalseEasting/2

        # assign grid to geometry
        LGRS_25km_geom_single_area = assign_xy_grid_to_multipolygon(Cylindrical_X,Cylindrical_Y)

        # collect global areas and merge 
        GlobalX,GlobalY,multigrid = generate_LGRS_grid("global",lonBand,h)
        # convert global zone boundary coordinates from latlon to LTM 
        for m in range(0,GlobalX.shape[0]):
            for n in range(0,GlobalY.shape[1]):
                if np.isnan(GlobalX[m,n]):
                    continue
                z,h,GlobalX[m,n],GlobalY[m,n] = toLTM(GlobalX[m,n],GlobalY[m,n],zone=None,trunc_val=0,process_errors=False)
        # assign to mutlipolygon
        LGRS_Global_geom = assign_arbitrary_xy_to_multipolygon(GlobalX,GlobalY)
        # union areas into 
        LGRS_25kmGlobal_geom_single_area = intersect_multipolygons(LGRS_25km_geom_single_area,LGRS_Global_geom)

        # determine LTM bounding box for clipping
        LTM_X,LTM_Y = generate_LTM_grid("border",lonBand,h)
        # convert LTM zone boundary coordinates from latlon to LTM 
        for m in range(0,LTM_X.shape[0]):
            for n in range(0,LTM_X.shape[1]):
                z,h,LTM_X[m,n],LTM_Y[m,n] = toLTM(LTM_X[m,n],LTM_Y[m,n],zone=None,trunc_val=0,process_errors=False)

        # clip 25km grid to LTM zone boundaries
        LTM_zone_clip_geom = assign_xy_to_polygon(LTM_X,LTM_Y)

        # clip to bounding box
        LGRS_area_clipped = clip_multipolygon_by_polygon(LGRS_25kmGlobal_geom_single_area,LTM_zone_clip_geom)

        # extract X, Y positions from clipped region
        X_vals,Y_vals = extract_xy_from_multipolygon(LGRS_area_clipped)

        # Assign final values to grid in LTM coordinates
        GridX[0:X_vals.shape[0],0:X_vals.shape[1]] = X_vals
        GridY[0:Y_vals.shape[0],0:Y_vals.shape[1]] = Y_vals
    
    # plotting smaller grids less then 25km resolution
    elif Grid_Res == "1km" or \
         Grid_Res == "100m" or \
         Grid_Res == "10m" or \
         Grid_Res == "1m":
        
        # determine grid size in meters not a string
        if Grid_Res == "1km":
            grid_sizeY,grid_sizeX = 1E3,1E3
        elif Grid_Res == "100m":
            grid_sizeY,grid_sizeX = 100,100
        elif Grid_Res == "10m":
            grid_sizeY,grid_sizeX = 10,10
        elif Grid_Res == "1m":
            grid_sizeY,grid_sizeX = 1,1

        if ulx == None or uly == None or lrx == None or lry == None:
            raise ValueError ("LGRS 1km grid requires an input box "
                              "provide ulx,uly,lrx,lry")

        # generate single zone grid: extend to clipping boarder.
        # for simplicity the grids are contained within a whole zone
        merge_grid = None
        # if merge grid is activated than the geometries of multiple
        # polygon geometries will be merged together in a union of
        # spaces. 

        # If the multigrid setting is not enabled check to make
        # sure 
        if not multigrid:
            if z1 != z2 or h1 != h2:
                choice = input("Map area crosses multiple zones "
                        "would you like export Coordinates in LatLong to "
                        "support multiple Zones? (y/n): ")
                
                # Double check that a multigrid plot is not wanted. 
                # Terminate if not
                if choice.lower() != 'y':
                    raise ValueError ("Plot Area crosses multiple zones: enable multigrid "
                                      "setting to plot this area")
                else:
                    multigrid = True

        # check if there is a wrap around the globe in a multigrid plot
        if z1>z2:
            # sting used to iterate though zones
            zone_range = list(range(z1,45+1,1))+list(range(1,z2+1,1))
            rev = True
        else:
            # normal zone range for multi plot. This works in one 
            # zone if z1 and z2 are the same
            zone_range = range(z1,z2+1,1)
            rev = False

        # iterate inducing and between z1 and z2
        for z in zone_range:
            
            # double check if the plot area crosses multiple hemispheres
            # and generate iteration list
            if h1 == h2:
                hems = [h1]
            else:
                hems = [h1,h2]

            # iterate through grid box hemispheres
            for h in hems: 
                LGRS_grid_clipped = None
                lam0 = Calc_LTM_prime_meridian(z,"deg")
                # create box for new Coordinates to cross zones

                # correct X position for multiple zones across the -180-180 meridian
                if rev and z == 1:
                    ztemp = z1
                    z2 = ztemp
                    z2 = z1

                # create left most corner of grid area
                if z == z1:
                    Cylindrical_Xmin = grid_sizeX*(ulx//grid_sizeX)
                else: # create left most area to confine to LTM zone
                    _,_,Cylindrical_Xmin,_ = toLTM(lam0-4,0,zone=None,trunc_val=0,process_errors=False) 
                    Cylindrical_Xmin = grid_sizeX*(Cylindrical_Xmin//grid_sizeX)

                # create righ portion of grid area
                if z == z2: 
                    Cylindrical_Xmax = grid_sizeX*(lrx//grid_sizeX+1)
                else: # create right most area to confine to LTM zone
                    _,_,Cylindrical_Xmax,_ = toLTM(lam0+4-1E-13,0,zone=None,trunc_val=0,process_errors=False) 
                    Cylindrical_Xmax = grid_sizeX*(Cylindrical_Xmax//grid_sizeX+1)
                
                # determine min and max of the grid area and if needed
                # process hemispheres for grid generation in both 
                # hemispheres
                if h1 == h2:
                    Cylindrical_Ymin = grid_sizeY*(lry//grid_sizeY)
                    Cylindrical_Ymax = grid_sizeY*(uly//grid_sizeY+1)

                elif h == "S": # format bottom of the grid
                    Cylindrical_Ymin = grid_sizeY*(lry//grid_sizeY+1)

                    Cylindrical_Ymax = FalseNorthing 
                    Cylindrical_Ymax = grid_sizeY*(Cylindrical_Ymax//grid_sizeY)

                elif h == "N": # format top of the grid
                    Cylindrical_Ymax = grid_sizeY*(uly//grid_sizeY+1)

                    Cylindrical_Ymin = 0 
                    Cylindrical_Ymin = grid_sizeY*(Cylindrical_Ymin//grid_sizeY)

                # Generate grid 
                GridY = np.arange(Cylindrical_Ymin,Cylindrical_Ymax,grid_sizeY)
                GridX = np.arange(Cylindrical_Xmin,Cylindrical_Xmax,grid_sizeX)

                # add in final endpoint to ensure correct positioning
                GridY = np.append(GridY,Cylindrical_Ymax)
                GridX = np.append(GridX,Cylindrical_Xmax)

                # create grid
                GridX,GridY = np.meshgrid(GridX,GridY)

                # convert LTM Coordinates to LatLon values
                # this supports a multigrid plot. It is reconverted later
                # if contained in a single zone
                for i in range(0,GridX.shape[0]):
                    for j in range(0,GridX.shape[1]):
                        GridX[i,j],GridY[i,j] = toLatLon(GridX[i,j],GridY[i,j],
                                                        zone=z,
                                                        h=h,
                                                        process_errors=False) 

                # assign values to polygon
                LGRS_grid_area = assign_xy_grid_to_multipolygon(GridX,GridY)
                # collect global areas and merge 
                
                GlobalX,GlobalY,_ = generate_LGRS_grid("global",z,h)

                # assign to mutlipolygon
                LGRS_Global_geom = assign_arbitrary_xy_to_multipolygon(GlobalX,GlobalY)
                # # union areas into 
                LGRS_AreaGlobal_geom_single_area = intersect_multipolygons(LGRS_grid_area,LGRS_Global_geom)

                # determine LTM bounding box for clipping
                LTM_X,LTM_Y = generate_LTM_grid("border",z,h)
                # print(LTM_X,LTM_Y)

                # assign to geometry for clipping
                LTM_zone_clip_geom = assign_xy_to_polygon(LTM_X,LTM_Y)

                # clip to bounding box geometry
                LGRS_grid_clipped = clip_multipolygon_by_polygon(LGRS_AreaGlobal_geom_single_area,LTM_zone_clip_geom)

                # processing flag to not call the merging function
                # on a multipolygon in the fist grid area
                if merge_grid == None:
                    merge_grid = LGRS_grid_clipped
                else:
                    # merge multiple polygons together. 
                    # output is the input on the next loop iteration.
                    merge_grid = merge_multipolygons(merge_grid,LGRS_grid_clipped)

        # print(LGRS_grid_area)

        # extract X, Y positions from clipped region
        GridX,GridY = extract_xy_from_multipolygon(merge_grid)

        # convert back to LTM Coordinates. if plot crosses multiple zones
        # leave in lat and lon
        if not multigrid:
            for i in range(0,GridX.shape[0]):
                for j in range(0,GridX.shape[1]):
                    if np.isnan(GridX[i,j]):
                        continue
                    _,_,GridX[i,j],GridY[i,j] = toLTM(GridX[i,j],GridY[i,j],zone=None,
                                                trunc_val=0,process_errors=False)
                        
    else:
        raise ValueError("LGRS Generation Terminated: Grid type could"
                         "not be determined. Please specify:\n"
                         "\{global,global_all,25km,25km_all,1km,100m, "
                         "10m, or 1m \}")

    # functionality to force the out put grid into lunar lat lon 
    # (not recomended as the grid will not be tied to a PCRS)
    if LatLon_override:
        if (not multigrid) and \
           ((Grid_Res != "global_all") and \
           (Grid_Res != "25km_all")  and \
           (Grid_Res != "global")):
            print(GridY)
            for i in range(0,GridX.shape[0]):
                for j in range(0,GridY.shape[1]):
                    if np.isnan(GridX[i,j]):
                        continue
                    GridX[i,j],GridY[i,j]=toLatLon(
                                            GridX[i,j],
                                            GridY[i,j], 
                                            z1,
                                            h, 
                                            process_errors=True)
        else:
            raise ValueError("Lat Lon Override could not be applied as "
                            "grids are already in a lat lon format:\n"
                            "Set latlon_override to False to continue")
    
    return GridX, GridY, multigrid

# ---------------------------------------------------------------------
# SECTION 2.18: Generate LGRS polar grid values according to a specified 
#              resolution and coordinate type


def generate_PolarLGRS_grid(Grid_Res,h=None,
                                h1=None,
                                ulx=None,
                                uly=None,
                                h2=None,
                                lrx=None,
                                lry=None,
                                LatLon_override=False):
    """
    This function is used to generate associated grids for the Lunar Grid
    Reference System (LGRS) in all resolutions and formats for the LPS 
    portions of the grid system.
    
    Input Perimeters:
    (int, float) Grid_Res - Grid format and resolution to be generated
    (str) h - LPS hemispheric location designator for LGRS plot for grids
              25km or greater in resolution.
    (str) h1 - LPS hemisphere designator for upper left hand coordinate
               of grid dimensioning bounding box. Only used in grid less-
               than or equal to 1km
    (double) ulx - upper left x coordinate of the grid dimensioning
                   bounding box. Only used in grids less than or equal
                   to 1km.
    (double) uly - upper left y coordinate of the grid dimensioning
                   bounding box. Only used in grids less than or equal
                   to 1km.
    (str) h2 - LPS hemisphere designator for lower right hand coordinate
               of grid dimensioning bounding box. Only used in grid less-
               than or equal to 1km
    (double) lrx - Lower right x coordinate of the grid dimensioning
                   bounding box. Only used in grids less than or equal
                   to 1km.
    (double) lry - Lower right y coordinate of the grid dimensioning
                   bounding box. Only used in grids less than or equal
                   to 1km.
    (bool) LatLon_override - Logical value that will force the inverse 
                             conversion of grid position back to lunar
                             latitude and longitude. the final exported 
                             grid will be be referenced to to lat lon.
                             NOTE: the potential for incorrect projections,
                             the LGRS and LGRS grids are meant to be 
                             utilized in the LPS and LTM PCRS. Without
                             utilization map grid distances will very and
                             map distortion will be calibrated. It is 
                             recommended that this option not be utilized 
                             to ensure correct positioning with the 
                             designed LGRS PCRS (LPS and LTM).

    Types of grids
    Grid_Res == global - Generate multipolygon grid of the global area
                         grid of the LGRS plot within a single specified
                         LPS zone. This is a two part plot of the global
                         easting value of the north pole and is plotted 
                         in LPS Coordinates in a stereographic projection.
    Grid_Res == global_all - Generate multipolygon grid of all global area
                         grids in LGRS. This plots across a single specified
                         LPS zone. This is a two part plot of the global
                         easting value of the north pole and is plotted 
                         in latlon Coordinates in a stereographic projection.
    Grid_Res == 25km - Generate multipolygon grid of all 25km grid
                        zones of PolarLGRS in a single LPS zone. 
                        This plots is calculated in the LPS coordinate format
                        with with the associated projection
    Grid_Res == 1km - Generate multipolygon 1km  PolarLGRS grid in 
                     a single LPS zones. Grid is created in LPS 
                     Coordinates in with the associated projection. 
                     Upper left and lower right corner point required.
    Grid_Res == 100m - Generate multipolygon 100m  LGRS grid in 
                     a single LPS zones. Grid is created in LPS 
                     Coordinates in with the associated projection. 
                     Upper left and lower right corner point required.
    Grid_Res == 10m - Generate multipolygon 10m  LGRS grid in 
                     a single LPS zones. Grid is created in LPS 
                     Coordinates in with the associated projection. 
                     Upper left and lower right corner point required.
    Grid_Res == 1m - Generate multipolygon 1m  LGRS grid in 
                     a single LPS zones. Grid is created in LPS 
                     Coordinates in with the associated projection. 
                     Upper left and lower right corner point required.
        Note: This function is used to calculate the output grids for
              LGRS associated coordinate formats inducing PolarLGRS, 
              PolarLGRS_ACC ( a full Artemis condensed format), and 
              PolarACC with global location truncation. 

    Returns:
    (2D array, floats) GridX - Longitude or Easting values of the the
                               polar LGRS grid in LPS coordinate. Grid
                               clipped within the LPS boarders.
    (2D array, floats) GridY - Latitude or Northing values of the the
                               polar LGRS grid in LPS coordinate. Grid
                               clipped within the LPS boarders.
    """
    # LGRS polar zones, in a single LPS zone LPS coordinates
    if Grid_Res == "global":

        if h ==  "S":
            # dimensions grid size by maximum possible size of the grid in X
            # for only two polygons
            GridX = np.full((2,int(2*(LPS_S_Boarder-LPS_S_phi0)/ResY+
                              int(LatLon_E_Meridian-LatLon_W_Meridian/ResX
                              +3))),np.nan)
            # dimensions grid size by maximum possible size of the grid in Y
            # for only two polygons
            GridY = np.full((2,int(2*(LPS_S_Boarder-LPS_S_phi0)/ResY+
                              int(LatLon_E_Meridian-LatLon_W_Meridian/ResX+
                              3))),np.nan)        
            
            # Polar LGRS Poly 1
            # Northern border
            GridX_B = np.arange(LatLon_W_Meridian,0,ResX)[:,np.newaxis] #grid lines
            GridX_B = np.append(GridX_B,0)

            GridY_B = np.ones(GridX_B.shape[0]-1).T*LPS_S_Boarder
            GridY_B = np.append(GridY_B,LPS_S_Boarder)

            # Crossing the hemisphere
            GridY_C1 = np.arange(LPS_S_Boarder,LPS_S_phi0,-ResY)[:,np.newaxis] #grid lines
            GridY_C1 = np.append(GridY_C1,LPS_S_phi0)

            GridX_C1 = np.ones(GridY_C1.shape[0]-1).T*0
            GridX_C1 = np.append(GridX_C1,0)
            

            GridY_C2 = np.arange(LPS_S_phi0,LPS_S_Boarder,ResY)[:,np.newaxis] #grid lines
            GridY_C2 = np.append(GridY_C2,LPS_S_Boarder)

            GridX_C2 = np.ones(GridY_C2.shape[0]-1).T*LatLon_W_Meridian
            GridX_C2 = np.append(GridX_C2,LatLon_W_Meridian)
            
            # combine grid values and concatenate into single list
            X_vals = (GridX_B,GridX_C1,GridX_C2)
            X_vals = np.concatenate(X_vals)

            Y_vals = (GridY_B,GridY_C1,GridY_C2)
            Y_vals = np.concatenate(Y_vals)
    
            # assign values to output arrays
            GridX[0,0:X_vals.shape[0]] = X_vals[:]
            GridY[0,0:Y_vals.shape[0]] = Y_vals[:]

            # Polar LGRS Poly 2
            GridX_B = np.arange(0,LatLon_E_Meridian,ResX)[:,np.newaxis] #grid lines
            GridX_B = np.append(GridX_B,LatLon_E_Meridian)

            GridY_B = np.ones(GridX_B.shape[0]-1).T*LPS_S_Boarder
            GridY_B = np.append(GridY_B,LPS_S_Boarder)

            # Crossing the hemisphere
            GridY_C1 = np.arange(LPS_S_phi0,LPS_S_Boarder,ResY)[:,np.newaxis] #grid lines
            GridY_C1 = np.append(GridY_C1,LPS_S_phi0)

            GridX_C1 = np.ones(GridY_C1.shape[0]-1).T*0
            GridX_C1 = np.append(GridX_C1,0)

            GridY_C2 = np.arange(LPS_S_Boarder,LPS_S_phi0,-ResY)[:,np.newaxis] #grid lines
            GridY_C2 = np.append(GridY_C2,LPS_S_phi0)

            GridX_C2 = np.ones(GridY_C2.shape[0]-1).T*LatLon_W_Meridian
            GridX_C2 = np.append(GridX_C2,0)
            
            # combine grid values and concatenate into single list
            X_vals = (GridX_B,GridX_C1,GridX_C2)
            X_vals = np.concatenate(X_vals)

            Y_vals = (GridY_B,GridY_C1,GridY_C2)
            Y_vals = np.concatenate(Y_vals)

            # assign values to output arrays
            GridX[1,0:X_vals.shape[0]] = X_vals[:]
            GridY[1,0:Y_vals.shape[0]] = Y_vals[:]

        elif h == "N":
            # dimensions grid size by maximum possible size of the grid in X
            # for only two polygons        
            GridX = np.full((2,int(2*(LPS_N_phi0-LPS_N_Boarder)/ResY+
                              int(LatLon_E_Meridian-LatLon_W_Meridian/ResX
                              +3))),np.nan)
            # dimensions grid size by maximum possible size of the grid in Y
            # for only two polygons
            GridY = np.full((2,int(2*(LPS_N_phi0-LPS_N_Boarder)/ResY+
                              int(LatLon_E_Meridian-LatLon_W_Meridian/ResX+
                              3))),np.nan)

            # Poly 1
            # Northern border
            GridX_B = np.arange(LatLon_W_Meridian,0,ResX)[:,np.newaxis] #grid lines
            GridX_B = np.append(GridX_B,0)

            GridY_B = np.ones(GridX_B.shape[0]-1).T*LPS_N_Boarder
            GridY_B = np.append(GridY_B,LPS_N_Boarder)

            # Crossing the hemisphere

            GridY_C1 = np.arange(LPS_N_phi0,LPS_N_Boarder,-ResY)[:,np.newaxis] #grid lines
            GridY_C1 = np.append(GridY_C1,LPS_N_Boarder)

            GridX_C1 = np.ones(GridY_C1.shape[0]-1).T*0
            GridX_C1 = np.append(GridX_C1,0)
            

            GridY_C2 = np.arange(LPS_N_Boarder,LPS_N_phi0,ResY)[:,np.newaxis] #grid lines
            GridY_C2 = np.append(GridY_C2,LPS_N_phi0)

            GridX_C2 = np.ones(GridY_C2.shape[0]-1).T*LatLon_W_Meridian
            GridX_C2 = np.append(GridX_C2,LatLon_W_Meridian)
            
            # combine grid values and concatenate into single list
            X_vals = (GridX_B,GridX_C1,GridX_C2)
            X_vals = np.concatenate(X_vals)
            Y_vals = (GridY_B,GridY_C1,GridY_C2)
            Y_vals = np.concatenate(Y_vals)
            
            # assign values to output arrays
            GridX[0,0:X_vals.shape[0]] = X_vals[:]
            GridY[0,0:Y_vals.shape[0]] = Y_vals[:]

            # grid zone 2
            GridX_B = np.arange(0,LatLon_E_Meridian,ResX)[:,np.newaxis] #grid lines
            GridX_B = np.append(GridX_B,LatLon_E_Meridian)

            GridY_B = np.ones(GridX_B.shape[0]-1).T*LPS_N_Boarder
            GridY_B = np.append(GridY_B,LPS_N_Boarder)

            # Crossing the hemisphere
            GridY_C1 = np.arange(LPS_N_Boarder,LPS_N_phi0,ResY)[:,np.newaxis] #grid lines
            GridY_C1 = np.append(GridY_C1,LPS_N_Boarder)

            GridX_C1 = np.ones(GridY_C1.shape[0]-1).T*0
            GridX_C1 = np.append(GridX_C1,0)
            

            GridY_C2 = np.arange(LPS_N_phi0,LPS_N_Boarder,-ResY)[:,np.newaxis] #grid lines
            GridY_C2 = np.append(GridY_C2,LPS_N_Boarder)

            GridX_C2 = np.ones(GridY_C2.shape[0]-1).T*LatLon_W_Meridian
            GridX_C2 = np.append(GridX_C2,0)

            # combine grid values and concatenate into single list
            X_vals = (GridX_B,GridX_C1,GridX_C2)
            X_vals = np.concatenate(X_vals)
            Y_vals = (GridY_B,GridY_C1,GridY_C2)
            Y_vals = np.concatenate(Y_vals)

            # assign values to output arrays
            GridX[1,0:X_vals.shape[0]] = X_vals[:]
            GridY[1,0:Y_vals.shape[0]] = Y_vals[:]

        # end processing if grid area could not be identified
        else:
            raise ValueError("LPS Global plot could not be plotted.")

    # LGRS polar zones in single LPS zone LatLon Coordinates
    elif Grid_Res == "global_all": 
        if h == "S":
            # dimensions grid size by maximum possible size of the grid in X
            # for only two polygons
            GridX = np.full((2,int(2*(LPS_S_Boarder-LPS_S_phi0)/ResY+
                              int(2*(LatLon_E_Meridian-LatLon_W_Meridian)/ResX
                              +4))),np.nan)
            # dimensions grid size by maximum possible size of the grid in Y
            # for only two polygons
            GridY = np.full((2,int(2*(LPS_S_Boarder-LPS_S_phi0)/ResY+
                              int(2*(LatLon_E_Meridian-LatLon_W_Meridian)/ResX+
                              4))),np.nan)  

            # Poly 1
            # L - > R border
            GridX_B = np.arange(LatLon_W_Meridian,0,ResX)[:,np.newaxis] #grid lines
            GridX_B = np.append(GridX_B,0)

            GridY_B = np.ones(GridX_B.shape[0]-1).T*LPS_S_phi0
            GridY_B = np.append(GridY_B,LPS_S_phi0)

            # S -> N
            GridY_R = np.arange(LPS_S_phi0,LPS_S_Boarder,ResY)[:,np.newaxis] #grid lines
            GridY_R = np.append(GridY_R,LPS_S_Boarder)

            GridX_R = np.ones(GridY_R.shape[0]-1).T*0
            GridX_R = np.append(GridX_R,0)
            
            # R -> L
            GridX_T = np.arange(0,LatLon_W_Meridian,-ResX)[:,np.newaxis] #grid lines
            GridX_T = np.append(GridX_T,LatLon_W_Meridian)

            GridY_T = np.ones(GridX_T.shape[0]-1).T*LPS_S_Boarder
            GridY_T = np.append(GridY_T,LPS_S_Boarder)
            
            # N -> S
            GridY_L = np.arange(LPS_S_Boarder,LPS_S_phi0,-ResY)[:,np.newaxis] #grid lines
            GridY_L = np.append(GridY_L,LPS_S_phi0)

            GridX_L = np.ones(GridY_L.shape[0]-1).T*LatLon_W_Meridian
            GridX_L = np.append(GridX_L,LatLon_W_Meridian)

            # combine parts of the polygon into list
            X_vals = (GridX_B,GridX_R,GridX_T,GridX_L)
            Y_vals = (GridY_B,GridY_R,GridY_T,GridY_L)

            # combine polygon into a single list
            X_vals = np.concatenate(X_vals)
            Y_vals = np.concatenate(Y_vals)

            # assigning grid to output arrays
            GridX[0,0:X_vals.shape[0]] = X_vals[:]
            GridY[0,0:Y_vals.shape[0]] = Y_vals[:]


            # Poly 2
            GridX_B = np.arange(0,LatLon_E_Meridian,ResX)[:,np.newaxis] #grid lines
            GridX_B = np.append(GridX_B,LatLon_E_Meridian)

            GridY_B = np.ones(GridX_B.shape[0]-1).T*LPS_S_phi0
            GridY_B = np.append(GridY_B,LPS_S_phi0)

            # S -> N
            GridY_R = np.arange(LPS_S_phi0,LPS_S_Boarder,ResY)[:,np.newaxis] #grid lines
            GridY_R = np.append(GridY_R,LPS_S_Boarder)

            GridX_R = np.ones(GridY_R.shape[0]-1).T*LatLon_E_Meridian
            GridX_R = np.append(GridX_R,LatLon_E_Meridian)
            
            # R -> L
            GridX_T = np.arange(LatLon_E_Meridian,0,-ResX)[:,np.newaxis] #grid lines
            GridX_T = np.append(GridX_T,LatLon_W_Meridian)

            GridY_T = np.ones(GridX_T.shape[0]-1).T*LPS_S_Boarder
            GridY_T = np.append(GridY_T,LPS_S_Boarder)
            
            # N -> S
            GridY_L = np.arange(LPS_S_Boarder,LPS_S_phi0,-ResY)[:,np.newaxis] #grid lines
            GridY_L = np.append(GridY_L,LPS_S_phi0)

            GridX_L = np.ones(GridY_L.shape[0]-1).T*0
            GridX_L = np.append(GridX_L,0)
            # combine parts of the polygon into list
            X_vals = (GridX_B,GridX_R,GridX_T,GridX_L)
            Y_vals = (GridY_B,GridY_R,GridY_T,GridY_L)

            # combine polygon into a single list
            X_vals = np.concatenate(X_vals)
            Y_vals = np.concatenate(Y_vals)

            # assigning grid to output arrays
            GridX[1,0:X_vals.shape[0]] = X_vals[:]
            GridY[1,0:Y_vals.shape[0]] = Y_vals[:]


        elif h == "N":
            # dimensions grid size by maximum possible size of the grid in X
            # for only two polygons
            GridX = np.full((2,int(2*(LPS_S_phi0-LPS_S_Boarder)/ResY+
                    int(2*(LatLon_E_Meridian-LatLon_W_Meridian)/ResX
                    +4))),np.nan)
            # dimensions grid size by maximum possible size of the grid in Y
            # for only two polygons
            GridY = np.full((2,int(2*(LPS_S_phi0-LPS_S_Boarder)/ResY+
                              int(2*(LatLon_E_Meridian-LatLon_W_Meridian)/ResX+
                              4))),np.nan)  
            # Poly 1
            # Northern border
        
            # Poly 1
            # L - > R border
            GridX_B = np.arange(LatLon_W_Meridian,0,ResX)[:,np.newaxis] #grid lines
            GridX_B = np.append(GridX_B,0)

            GridY_B = np.ones(GridX_B.shape[0]-1).T*LPS_N_Boarder
            GridY_B = np.append(GridY_B,LPS_N_Boarder)

            # S -> N
            GridY_R = np.arange(LPS_N_Boarder,LPS_N_phi0,ResY)[:,np.newaxis] #grid lines
            GridY_R = np.append(GridY_R,LPS_N_phi0)

            GridX_R = np.ones(GridY_R.shape[0]-1).T*0
            GridX_R = np.append(GridX_R,0)
            
            # R -> L
            GridX_T = np.arange(0,LatLon_W_Meridian,-ResX)[:,np.newaxis] #grid lines
            GridX_T = np.append(GridX_T,LatLon_W_Meridian)

            GridY_T = np.ones(GridX_T.shape[0]-1).T*LPS_N_phi0
            GridY_T = np.append(GridY_T,LPS_N_phi0)
            
            # N -> S
            GridY_L = np.arange(LPS_N_phi0,LPS_N_Boarder,-ResY)[:,np.newaxis] #grid lines
            GridY_L = np.append(GridY_L,LPS_N_Boarder)

            GridX_L = np.ones(GridY_L.shape[0]-1).T*LatLon_W_Meridian
            GridX_L = np.append(GridX_L,LatLon_W_Meridian)

            # combine parts of polygon into list
            X_vals = (GridX_B,GridX_R,GridX_T,GridX_L)
            Y_vals = (GridY_B,GridY_R,GridY_T,GridY_L)

            # concatenate values into single array
            X_vals = np.concatenate(X_vals)
            Y_vals = np.concatenate(Y_vals) 

            # combine grid values
            GridX[0,0:X_vals.shape[0]] = X_vals[:]
            GridY[0,0:Y_vals.shape[0]] = Y_vals[:]

            # Poly 2
            GridX_B = np.arange(0,LatLon_E_Meridian,ResX)[:,np.newaxis] #grid lines
            GridX_B = np.append(GridX_B,LatLon_E_Meridian)

            GridY_B = np.ones(GridX_B.shape[0]-1).T*LPS_N_Boarder
            GridY_B = np.append(GridY_B,LPS_N_Boarder)

            # S -> N
            GridY_R = np.arange(LPS_N_Boarder,LPS_N_phi0,ResY)[:,np.newaxis] #grid lines
            GridY_R = np.append(GridY_R,LPS_N_phi0)

            GridX_R = np.ones(GridY_R.shape[0]-1).T*LatLon_E_Meridian
            GridX_R = np.append(GridX_R,LatLon_E_Meridian)
            
            # R -> L
            GridX_T = np.arange(LatLon_E_Meridian,0,-ResX)[:,np.newaxis] #grid lines
            GridX_T = np.append(GridX_T,LatLon_W_Meridian)

            GridY_T = np.ones(GridX_T.shape[0]-1).T*LPS_N_phi0
            GridY_T = np.append(GridY_T,LPS_N_phi0)
            
            # N -> S
            GridY_L = np.arange(LPS_N_phi0,LPS_N_Boarder,-ResY)[:,np.newaxis] #grid lines
            GridY_L = np.append(GridY_L,LPS_N_Boarder)

            GridX_L = np.ones(GridY_L.shape[0]-1).T*0
            GridX_L = np.append(GridX_L,0)

            # combine grid values into a list
            X_vals = (GridX_B,GridX_R,GridX_T,GridX_L)
            X_vals = np.concatenate(X_vals)
            Y_vals = (GridY_B,GridY_R,GridY_T,GridY_L)
            Y_vals = np.concatenate(Y_vals) 

            # assign grid values
            GridX[1,0:X_vals.shape[0]] = X_vals[:]
            GridY[1,0:Y_vals.shape[0]] = Y_vals[:]

        else: 
            raise ValueError ("Hemisphere could not be determined for polar"
                              "LGRS global all grid.")

    #  25km grid areas, in a single LPS zone in LPS Coordinates
    elif Grid_Res == "25km": 
        if h == "S":
            # generate empty arrays for processing 
            GridX = np.full((2,int(2*(LPS_S_Boarder-LPS_S_phi0)/ResY+
                            int(LatLon_E_Meridian-LatLon_W_Meridian/ResX
                            +3))),np.nan)

            GridY = np.full((2,int(2*(LPS_S_Boarder-LPS_S_phi0)/ResY+
                              int(LatLon_E_Meridian-LatLon_W_Meridian/ResX+
                              3))),np.nan) 
            
            # Determine LPS zone max and min values in meters
            # these are quadrant values, azimuth on the LPS
            # plain, not on the sphere.
            h,X_max,Y = toLPS(90,LPS_S_Boarder,trunc_val=0)
            h,X_min,Y = toLPS(-90,LPS_S_Boarder,trunc_val=0)
            h,X,Y_max = toLPS(-.0,LPS_S_Boarder,trunc_val=0)
            h,X,Y_min = toLPS(180,LPS_S_Boarder,trunc_val=0)

        elif h == "N":
            # generate empty arrays for processing 
            GridX = np.full((2,int(2*(LPS_N_phi0-LPS_N_Boarder)/ResY+
                            int(LatLon_E_Meridian-LatLon_W_Meridian/ResX
                            +3))),np.nan)

            GridY = np.full((2,int(2*(LPS_N_phi0-LPS_N_Boarder)/ResY+
                              int(LatLon_E_Meridian-LatLon_W_Meridian/ResX+
                              3))),np.nan) 

            # Determine LPS zone max and min values in meters
            # these are quadrant values, azimuth on the LPS
            # plain, not on the sphere.
            h,X_max,Y = toLPS(90,LPS_N_Boarder,trunc_val=0)
            h,X_min,Y = toLPS(-90,LPS_N_Boarder,trunc_val=0)
            h,X,Y_max = toLPS(180,LPS_N_Boarder,trunc_val=0)
            h,X,Y_min = toLPS(-.0,LPS_N_Boarder,trunc_val=0)

        else:   
            raise ValueError ("LPS zone for Polar LGRS grid could not be determined")
        
        # generate arrays. These values are in meters for the stereographic projection
        # get min and max of grid zones to be generated
        GridY_min = Y_min//25E3*25E3
        GridY_max = (Y_max//25E3 + 1)*25E3

        GridX_min = X_min//25E3*25E3
        GridX_max = (X_max//25E3 + 1)*25E3

        # generate 25km grid
        StereoY = np.arange(np.abs(GridY_min),np.abs(GridY_max)+25E3,25E3)
        StereoX = np.arange(np.abs(GridX_min),np.abs(GridX_max)+25E3,25E3)

        # meshgrid for array of Xs and Ys
        StereoY,StereoX = np.meshgrid(StereoX,StereoY)

        # assign grid to geometry
        LGRS_25km_geom_polar_area = assign_xy_grid_to_multipolygon(StereoY,StereoX)

        # determine LTM bounding box for clipping
        LPS_X,LPS_Y = generate_LPS_grid("global",h)

        # convert coordinates to LPS values
        for m in range(0,LPS_X.shape[0]):
            for n in range(0,LPS_X.shape[1]):
                _,LPS_X[m,n],LPS_Y[m,n] = toLPS(LPS_X[m,n],LPS_Y[m,n],trunc_val=0,process_errors=False)

        # Assign LPS zone boarder to a polygon geometry             
        LPS_zone_clip_geom = assign_xy_to_polygon(LPS_X,LPS_Y)

        # clip to bounding box
        LGRS_area_clipped = clip_multipolygon_by_polygon(LGRS_25km_geom_polar_area,LPS_zone_clip_geom)

        # extract X, Y positions from clipped region
        GridX,GridY = extract_xy_from_multipolygon(LGRS_area_clipped)

    # generate polar grid in grid sizes less then 25km
    elif Grid_Res == "1km" or \
         Grid_Res == "100m" or \
         Grid_Res == "10m" or \
         Grid_Res == "1m":

        # making a grid in multiple zones is not allowed for LGRS in
        # the LPS portions. Terminate if crosses into both polar 
        # regions
        if h1!=h2:
            raise ValueError("Polar LGRS cannot be generated in both"
                             "at the same time:\n Please double check"
                             "coordinates polar values")

        # get grid dimensions in meters
        if Grid_Res == "1km":
            grid_sizeY,grid_sizeX = 1E3,1E3
        elif Grid_Res == "100m":
            grid_sizeY,grid_sizeX = 100,100
        elif Grid_Res == "10m":
            grid_sizeY,grid_sizeX = 10,10
        elif Grid_Res == "1m":
            grid_sizeY,grid_sizeX = 1,1

        if ulx == None or uly == None or lrx == None or lry == None:
            raise ValueError ("LGRS <1km polar grid requires an input box "
                              "provide ulx,uly,lrx,lry")
        
        # generate single zone grid: extend to clipping boarder.
        # for simplicity the grids are contained within a whole zone

        # create box for new Coordinates to cross zones
        StereoX_min = grid_sizeX*(ulx//grid_sizeX)
        StereoX_max = grid_sizeX*(lrx//grid_sizeX+1)

        StereoY_min = grid_sizeY*(lry//grid_sizeY)
        StereoY_max = grid_sizeY*(uly//grid_sizeY + 1)

        # Generate arrays to specified size
        GridY = np.arange(StereoY_min,StereoY_max,grid_sizeY)
        GridX = np.arange(StereoX_min,StereoX_max,grid_sizeX)

        # append last value to guarantee endpoint
        GridY = np.append(GridY,StereoY_max)
        GridX = np.append(GridX,StereoX_max)

        # create grid
        GridX,GridY = np.meshgrid(GridX,GridY)

        # testing function to see if the clipping function is
        # needed. Clipping in polar regions takes a longtime
        # at high fidelity grids due to the number of polygons
        clip = False
        for i in range(0,GridY.shape[0]):
            for j in range(0,GridY.shape[1]):
                _,Y = toLatLon_Polar(GridX[i,j],GridY[i,j],h1,
                            eqs="Spherical",process_errors=False)
                # search to see if any point is outside of the LPS plot
                # area. Set clip to true if true

                if ( (h1 == "N" and Y < 80) or\
                    (h1 == "S" and Y > -80) ):
                    clip = True
                    break

        # is a clipping operation is needed to be applied to the
        # Polar LGRS region
        if clip:
            # assign grid LPS coordinates to polygon
            LGRS_grid_area = assign_xy_grid_to_multipolygon(GridX,GridY)
            
            # determine LPS bounding box for clipping
            LPS_X,LPS_Y = generate_LPS_grid("border",h1)

            # convert LPS boarder to LPS grid Coordinates
            for i in range(0,LPS_X.shape[1]):
                _,LPS_X[0,i],LPS_Y[0,i] = toLPS(LPS_X[0,i],LPS_Y[0,i],trunc_val=0,
                                    eqs="Spherical",process_errors=True)
            
            # assign to geometry for clipping
            LPS_zone_clip_geom = assign_xy_to_polygon(LPS_X,LPS_Y)

            # clip to bounding box
            LGRS_grid_clipped = clip_multipolygon_by_polygon(LGRS_grid_area,LPS_zone_clip_geom)

            # extract X, Y positions from clipped region
            GridX,GridY = extract_xy_from_multipolygon(LGRS_grid_clipped)

        # if clipping is not needed
        else:
            # assign grid area to polygon
            LGRS_grid_area = assign_xy_grid_to_multipolygon(GridX,GridY)

            # export X and Y values from grid. This is completed for 
            # proper formatting
            GridX,GridY = extract_xy_from_multipolygon(LGRS_grid_area)
    else:
        raise ValueError("PolarLGRS Generation Terminated: Grid type could"
                    "not be determined. Please specify:\n"
                    "\{global,global_all,25km,25km_all,1km,100m, "
                    "10m, or 1m \}")
    
    # functionality to force the out put grid into lunar lat lon 
    # (not recomended as the grid will not be tied to a PCRS)
    print(LatLon_override)

    if LatLon_override:
        if Grid_Res != "global_all":
            for i in range(0,GridX.shape[0]):
                for j in range(0,GridY.shape[1]):
                    GridX[i,j],GridY[i,j]=toLatLon_Polar(GridX[i,j],
                                                        GridY[i,j], 
                                                        h1,
                                                        eqs="Spherical",
                                                        process_errors=True)
        else:
            raise ValueError("Lat Lon Override could not be applied as "
                            "grids are already in a lat lon format:\n"
                            "Set latlon_override to False to continue")    
    return GridX, GridY

# =====================================================================
# SECTION 3: MAIN PROGRAM 

# SECTION 3.0: Initialize and run main program
# SECTION 3.1: determine if input arguments are correct
# SECTION 3.2:  Read in input arguments center x center y or corners,
# SECTION 3.3: Converting Input Coordinates to Common Grid Format.
# SECTION 3.4: Generate the Correct Grid to the desired resolution
# SECTION 3.5: Convert grid Coordinates to correct output format.
# SECTION 3.6: Determine grid cell names in specified output format
# SECTION 3.7: Assign output WKT1 Projection string for writing file
# SECTION 3.8:  Write grid to ESRI shape file and format output names

# ---------------------------------------------------------------------
# SECTION 3.0: Initialize and run main program

def main(form_in, form_out, Grid_Res,
                            multigrid=False,
                            attribute=True,
                            overwrite=False,
                            output_file_form="shp",
                            info=True,
                            LatLon_override=False):
    
    """
    Run main program and generate arrays. Descriptions 
    provided for each step

    Input Perimeters:
    (str) form_in - Coordinate input format
    (str) form_out - Coordinate output format
    (str) Grid_Res - type of grid to be generated
    (str) units - specify unites (placed for later code improvements)
    (bool) multigrid - processing flag that enables an LTM grid to 
                    be generated across multiple zones and hemispheres.
                    grid is forced in latlon
    (bool) attribute - processing flag that enables export of a 
                    a shapefiles attribute table
    (bool) overwrite - processing flag that will remover a shape file if
                    it already exists.
    (str) output_file_form - output file format. Current support 
                            induces esri shapefiles and geopackages 
    (bool) info - processing flag prints processing information to the
                    user.
    (bool) LatLon_override - Force the grids to output in lunar latlons
                        only applies to LGRS and LGRS in ACC format
                        NOTE: these projections must still be used
                        in LPS or LTM to ensure correct spatial
                        dimensions.

    Correct format in values for form_in
    Lunar Transverse Mercator - (LTM)
    Lunar Polar Stereographic - (LPS)
    Lunar grid Reference System - (LGRS)
    Polar LGRS - (PolarLGRS)
    LGRS in Artemis Condensed Coordinate (ACC) Format - (LGRS_ACC)
    Polar LGRS in ACC Format - (PolarLGRS_ACC)

    Correct format out values for form_out
    LTM
    LPS
    LGRS
    LGRS_ACC
    ACC
    PolarLGRS
    PolarLGRS_ACC
    LGRS in ACC Format truncated to 10m without global 
                    reference - (PolarLGRS_ACC)
    Polar LGRS in ACC Format truncated to 10m without global 
                    reference - (PolarACC)

    Types of grids for Grid_Res
    global_all - All global areas 
    global - global areas within a single LTM or LPS zone
    border - LTM or LPS zone boarders
    25km - LGRS rectangular grid at a 25km resolution
    25km_all - Generate all LGRS 25km grid areas
    1km - LGRS rectangular grid at a 1km resolution
    100m - LGRS rectangular grid at a 100m resolution
    10m - LGRS rectangular grid at a 10m resolution
    1m - LGRS rectangular grid at a 1m resolution

    Returns:
    None

    Exports:
    (file) - esri shapefile (shp,shx,dbf, and prj) 
    (folder) - esri shape file containing folder 
    """

    # using Coordinate Conversion program, initialize its Globals
    # for calculations
    initialize_LGRS_function_globals()

# ---------------------------------------------------------------------
# SECTION 3.1: determine if input arguments are correct
#       grid and grid resolution are checked. 
#       We have 8 different coordinate formats and 7 different grid sizes
#       Not all work together. This section determines if GridRes and
#       form_in are comparable.

    
    try:
#       Acceptable coordinate system and grid resolution combinations
        if not ( ( form_out == "LTM" and Grid_Res == "global_all" ) or \
             ( form_out == "LTM" and Grid_Res == "global" ) or \
             ( form_out == "LTM" and Grid_Res == "border" ) or \
             ( form_out == "LPS" and Grid_Res == "global" ) or \
             ( form_out == "LPS" and Grid_Res == "border" ) or \
             ( form_out == "LGRS" and Grid_Res == "global" ) or \
             ( form_out == "LGRS" and Grid_Res == "global_all" ) or \
             ( form_out == "LGRS" and Grid_Res == "border" ) or \
             ( form_out == "LGRS" and Grid_Res == "25km" ) or \
             ( form_out == "LGRS" and Grid_Res == "25km_all" ) or \
             ( form_out == "LGRS" and Grid_Res == "1km" ) or \
             ( form_out == "LGRS" and Grid_Res == "100m" ) or \
             ( form_out == "LGRS" and Grid_Res == "10m" ) or \
             ( form_out == "LGRS" and Grid_Res == "1m" ) or \
             ( form_out == "PolarLGRS" and Grid_Res == "global" ) or \
             ( form_out == "PolarLGRS" and Grid_Res == "global_all" ) or \
             ( form_out == "PolarLGRS" and Grid_Res == "25km" ) or \
             ( form_out == "PolarLGRS" and Grid_Res == "1km" ) or \
             ( form_out == "PolarLGRS" and Grid_Res == "100m" ) or \
             ( form_out == "PolarLGRS" and Grid_Res == "10m" ) or \
             ( form_out == "PolarLGRS" and Grid_Res == "1m" ) or \
             ( form_out == "LGRS_ACC" and Grid_Res == "1km" ) or \
             ( form_out == "LGRS_ACC" and Grid_Res == "100m" ) or \
             ( form_out == "LGRS_ACC" and Grid_Res == "10m" ) or \
             ( form_out == "LGRS_ACC" and Grid_Res == "1m" ) or \
             ( form_out == "PolarLGRS_ACC" and Grid_Res == "1km" ) or \
             ( form_out == "PolarLGRS_ACC" and Grid_Res == "100m" ) or \
             ( form_out == "PolarLGRS_ACC" and Grid_Res == "10m" ) or \
             ( form_out == "PolarLGRS_ACC" and Grid_Res == "1m" ) or \
             ( form_out == "ACC" and Grid_Res == "1km" ) or \
             ( form_out == "ACC" and Grid_Res == "100m" ) or \
             ( form_out == "ACC" and Grid_Res == "10m" ) or \
             ( form_out == "PolarACC" and Grid_Res == "1km" ) or \
             ( form_out == "PolarACC" and Grid_Res == "100m" ) or \
             ( form_out == "PolarACC" and Grid_Res == "10m" ) ):
            raise ValueError("Output coordinate system and Grid Resolution cannot "
                            "be plotted. Please provide on of the following:\n"
                            "Please provide one of the following:\n:"
                            "form_out: LTM | grid_res: global_all or,\n"
                            "form_out: LTM | grid_res: global or, \n"
                            "form_out: LTM | grid_res: border or,\n"
                            "form_out: LPS | grid_res: border or,\n"
                            "form_out: LPS | grid_res: global or,\n"
                            "form_out: LGRS | grid_res: global or,\n"
                            "form_out: LGRS | grid_res: global_all or,\n"
                            "form_out: LGRS | grid_res: 25km or,\n"
                            "form_out: LGRS | grid_res: 25km_all or,\n"
                            "form_out: LGRS | grid_res: 1km or,\n"
                            "form_out: LGRS | grid_res: 100m or,\n"
                            "form_out: LGRS | grid_res: 10m or,\n"
                            "form_out: LGRS | grid_res: 1m or,\n"
                            "form_out: PolarLGRS | grid_res: global or,\n"
                            "form_out: PolarLGRS | grid_res: global_all or,\n"
                            "form_out: PolarLGRS | grid_res: 25km or,\n"
                            "form_out: PolarLGRS | grid_res: 1km or,\n"
                            "form_out: PolarLGRS | grid_res: 100m or,\n"
                            "form_out: PolarLGRS | grid_res: 10m or,\n"
                            "form_out: PolarLGRS | grid_res: 1m or,\n"
                            "form_out: LGRS_ACC | grid_res: 1km or,\n"
                            "form_out: LGRS_ACC | grid_res: 100m or,\n"
                            "form_out: LGRS_ACC | grid_res: 10m or,\n"
                            "form_out: LGRS_ACC | grid_res: 1m or,\n"
                            "form_out: PolarLGRS_ACC | grid_res: 1km or,\n"
                            "form_out: PolarLGRS_ACC | grid_res: 100m or,\n"
                            "form_out: PolarLGRS_ACC | grid_res: 10m or,\n"
                            "form_out: PolarLGRS_ACC | grid_res: 1m  or,\n"
                            "form_out: ACC | Grid_Res: 1km, or \n"
                            "form_out: ACC | Grid_Res: 100m, or \n"
                            "form_out: ACC | Grid_Res: 10m, or \n"
                            "form_out: PolarACC | Grid_Res: 1km or, \n"
                            "form_out: PolarACC | Grid_Res: 100m or, \n"
                            "form_out: PolarACC | Grid_Res: 10m")
        
        # this checks to make sure the right input coordinates can
        # be used to create a grid
        if not ( ( form_in == "LTM" and form_out == "LTM" ) or \
             ( form_in == "LTM" and form_out == "LGRS" ) or \
             ( form_in == "LTM" and form_out == "LGRS_ACC" ) or \
             ( form_in == "LTM" and form_out == "ACC" ) or \
             ( form_in == "LPS" and form_out == "LPS" ) or \
             ( form_in == "LPS" and form_out == "PolarLGRS" ) or \
             ( form_in == "LPS" and form_out == "PolarLGRS_ACC" ) or \
             ( form_in == "LPS" and form_out == "PolarACC" ) or \
             ( form_in == "LGRS" and form_out == "LGRS" ) or \
             ( form_in == "LGRS" and form_out == "LGRS_ACC" ) or \
             ( form_in == "LGRS" and form_out == "ACC" ) or \
             ( form_in == "PolarLGRS" and form_out == "PolarLGRS" ) or \
             ( form_in == "PolarLGRS" and form_out == "PolarLGRS_ACC" ) or \
             ( form_in == "PolarLGRS" and form_out == "PolarACC" ) or \
             ( form_in == "LGRS_ACC" and form_out == "LGRS_ACC" ) or \
             ( form_in == "LGRS_ACC" and form_out == "ACC" ) or \
             ( form_in == "PolarLGRS_ACC" and form_out == "PolarLGRS_ACC" ) or \
             ( form_in == "PolarLGRS_ACC" and form_out == "PolarACC" ) or \
             ( form_in == "LatLon" and form_out == "LTM" ) or \
             ( form_in == "LatLon" and form_out == "LGRS" ) or \
             ( form_in == "LatLon" and form_out == "LGRS_ACC" ) or \
             ( form_in == "LatLon" and form_out == "ACC" ) or \
             ( form_in == "PolarLatLon" and form_out == "LPS" ) or \
             ( form_in == "PolarLatLon" and form_out == "PolarLGRS" ) or \
             ( form_in == "PolarLatLon" and form_out == "PolarLGRS_ACC" ) or \
             ( form_in == "PolarLatLon" and form_out == "PolarACC" ) ):           
            raise ValueError("Input coordinate system and Output coordinate system "
                    "cannot be be processed. Please provide on of the following:\n"
                    "form_in: LTM | form_out: LTM, or\n"
                    "form_in: LTM | form_out: LGRS, or\n"
                    "form_in: LTM | form_out: LGRS_ACC, or\n"
                    "form_in: LTM | form_out: ACC, or\n"
                    "form_in: LPS | form_out: LPS, or\n"
                    "form_in: LPS | form_out: PolarLGRS, or\n"
                    "form_in: LPS | form_out: PolarLGRS_ACC, or\n"
                    "form_in: LPS | form_out: PolarACC, or\n"
                    "form_in: LGRS | form_out: LGRS, or\n"
                    "form_in: LGRS | form_out: LGRS_ACC, or\n"
                    "form_in: LGRS | form_out: ACC, or\n"
                    "form_in: PolarLGRS | form_out: PolarLGRS, or\n"
                    "form_in: PolarLGRS | form_out: PolarLGRS_ACC, or\n"
                    "form_in: PolarLGRS | form_out: PolarACC, or\n"
                    "form_in: LGRS_ACC | form_out: LGRS_ACC, or\n"
                    "form_in: LGRS_ACC | form_out: ACC, or\n"
                    "form_in: PolarLGRS_ACC | form_out: PolarLGRS_ACC, or\n"
                    "form_in: PolarLGRS_ACC | form_out: PolarACC, or\n"
                    "form_in: LatLon | form_out: LTM, or\n"
                    "form_in: LatLon | form_out: LGRS, or\n"
                    "form_in: LatLon | form_out: LGRS_ACC, or\n"
                    "form_in: LatLon | form_out: ACC, or\n"
                    "form_in: PolarLatLon | form_out: LPS, or\n"
                    "form_in: PolarLatLon | form_out: PolarLGRS, or\n"
                    "form_in: PolarLatLon | form_out: PolarLGRS_ACC, or\n"
                    "form_in: PolarLatLon | form_out: PolarACCPolarACC.")
    # print faults
    except ValueError as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        print(exc_type, fname, exc_tb.tb_lineno)
        print(e)
        exit(1)

# ---------------------------------------------------------------------
# SECTION 3.2:  Read in input arguments center x center y or corners,
#                (ulx uly lrx lry) to form a bounding  box to dimension 
#               the grid in


    try:
        # Planetocentric Latitude and Longitude
        if form_in == "LatLon":
            if Grid_Res == "global_all" or Grid_Res == "25km_all" :
                h,z = None,None
            elif Grid_Res not in grid_format[5:]: # 25km or greater
                if len(sys.argv)-1 == 5:
                    cent_x=float(sys.argv[4])
                    cent_y=float(sys.argv[5])
                else:
                    raise IndexError("Operation aborted: Number of inputs not correct."
                                    "for grid >=25km: coordinate: Center_X Center_Y")
            elif Grid_Res in grid_format:
                if len(sys.argv)-1 == 7:
                    ulx = float(sys.argv[4])
                    uly = float(sys.argv[5])
                    lrx = float(sys.argv[6])
                    lry = float(sys.argv[7])
                else:
                    raise IndexError("Operation aborted: Number of inputs not correct."
                        "for grid <25km LatLon:\nulx Lon\nuly lat\nlrx lon\nlry lat")
            else: 
                raise IndexError("Operation aborted: Number of inputs not correct."
                        "Input for LatLon:\nulx Lat\nuly Lon\nlrx Lat\nlry Lon"
                        "or the single coordinate: Center_X Center_Y")
            
        elif form_in == "PolarLatLon":
            if Grid_Res not in grid_format[5:]: # 25km or greater
                if len(sys.argv)-1 == 5:
                    cent_x = float(sys.argv[4])
                    cent_y = float(sys.argv[5])
                else:
                    raise IndexError("Operation aborted: Number of inputs not correct."
                                    "for grid >=25km: coordinate: Center_X Center_Y")
            elif Grid_Res in grid_format:
                if len(sys.argv)-1 == 7:
                    ulx = float(sys.argv[4])
                    uly = float(sys.argv[5])
                    lrx = float(sys.argv[6])
                    lry = float(sys.argv[7])
                else:
                    raise IndexError("Operation aborted: Number of inputs not correct."
                        "for grid <25km LatLon:\nulx Lon\nuly lat\nlrx lon\nlry lat")
            else: 
                raise IndexError("Operation aborted: Number of inputs not correct."
                        "Input for LatLon:\nulx Lat\nuly Lon\nlrx Lat\nlry Lon"
                        "or the single coordinate: Center_X Center_Y")
        # Lunar Transverse Mercator  
        elif form_in == "LTM":
            if Grid_Res == "global_all" or Grid_Res == "25km_all" :
                h,z = None,None
            elif Grid_Res not in grid_format[5:]:
                if len(sys.argv)-1 == 5:
                    z = int(sys.argv[4])
                    h = str(sys.argv[5])

                elif len(sys.argv)-1 == 4:
                    try:
                        z,h = list(re.findall(LTM_pattern,sys.argv[4])[0])
                        z = int(z)
                        h = str(h)

                    except:
                        raise IndexError("Operation aborted: Number of inputs not correct."
                            "\nInput for LTM: zone\{1-45\}, hemisphere \"S\" | \"N\"")
                else: 
                    raise IndexError("Operation aborted: Number of inputs not correct."
                                    "\nInput for LTM: zone\{1-45\}, hemisphere \"S\" | \"N\"")
            elif Grid_Res in grid_format[5:]:
                    if len(sys.argv)-1 == 5:
                        try:
                            z1,h1,ulx,_,uly,_ = list(re.findall(LTM_pattern_full,sys.argv[4])[0])
                            z2,h2,lrx,_,lry,_ = list(re.findall(LTM_pattern_full,sys.argv[5])[0])

                            z1,h1,ulx,uly = int(z1),str(h1),float(ulx),float(uly)
                            z2,h2,lrx,lry = int(z2),str(h2),float(lrx),float(lry)
                        except: 
                            raise IndexError("Operation aborted: LTM Coordinates could not be"
                                          " converted to proper format correctly.")
                    else:
                        raise IndexError("Operation aborted: Number of inputs not correct."
                            "\nInput for LTM: zone1\{1-45\}, hemisphere1\"S\"| \"N\" Easting 1, Northing 1"
                            "\nInput for LTM: zone2\{1-45\}, hemisphere2\"S\"| \"N\" Easting 2, Northing 2")
            else:
                raise IndexError("Operation aborted: Input Coordinates could not be read")
            # convert to proper data formats

        # Lunar polar stereographic Coordinates
        elif form_in == "LPS":

            if Grid_Res not in grid_format[5:]:
                if len(sys.argv)-1 == 4:
                    try:
                        h = str(sys.argv[4])
                    except: 
                        raise IndexError("Operation aborted: LPS Coordinates could not be"
                                " converted to proper format correctly.")
                                
            elif Grid_Res in grid_format[5:]:
                if len(sys.argv)-1 == 5:   
                    try:
                        h1,ulx,_,uly,_ = list(re.findall(LPS_pattern_full,sys.argv[4])[0])
                        h2,lrx,_,lry,_ = list(re.findall(LPS_pattern_full,sys.argv[5])[0])
                    except:
                        raise IndexError("Operation aborted: LPS Coordinates could not be"
                                " converted to proper format correctly. Check input arguments")
                        
                    h1,ulx,uly = str(h1),float(ulx),float(uly)
                    h2,lrx,lry = str(h2),float(lrx),float(lry)
                
                else: 
                    raise IndexError("Operation aborted: Number of inputs not correct."
                            "\nInput for LPS: hemisphere1\"S\"| \"N\" Easting 1, Northing 1"
                            "\nInput for LPS: hemisphere2\"S\"| \"N\" Easting 2, Northing 2")

        # Lunar Grid Reference System in LTM regions
        elif form_in == "LGRS":
            if Grid_Res  ==  "global_all" or Grid_Res == "25km_all":
                z,h = None,None

            elif Grid_Res not in grid_format[5:]: # 25km or greater
                if len(sys.argv)-1 == 5: #single zone area
                    z = sys.argv[4]
                    h = sys.argv[5]

                elif len(sys.argv)-1 == 4: # bounding box in condensed Coordinates
                    try:
                        z,h = list(re.findall(LTM_pattern,sys.argv[4])[0])

                    except:
                        
                        raise IndexError("Operation aborted: Number of inputs not correct."
                            "Input for LGRS coarse grid (>=25km):\n Zone \{1-45\}"
                            " hemisphere \"S\" | \"N\"")
                else: 
                    raise IndexError("Operation aborted: Number of inputs not correct."
                                    "Maybe try a smaller grid side.\n"
                                    "Input for LGRS coarse grid (>=25km):\n Zone \{1-45\}"
                                    " hemisphere \"S\" | \"N\"")
                # convert to proper data formats
                try:
                    z = int(z)
                    h = str(h)
                except: 
                    raise IndexError("Operation aborted: LGRS Coordinates could not be"
                                " converted to proper format correctly.")

            elif Grid_Res in grid_format:
                if len(sys.argv)-1 == 15:

                    lonBand1 = sys.argv[4]
                    latBand1 = sys.argv[5]
                    e25k1 = sys.argv[6]
                    n25k1 = sys.argv[7]
                    E1 = sys.argv[8]
                    N1 = sys.argv[9]

                    lonBand2 = sys.argv[10]
                    latBand2 = sys.argv[11]
                    e25k2 = sys.argv[12]
                    n25k2 = sys.argv[13]
                    E2 = sys.argv[14]
                    N2 = sys.argv[15]

                elif len(sys.argv)-1 == 5: # bounding box in condensed Coordinates
                    try:
                        lonBand1,latBand1,e25k1,n25k1,E1,N1=list(re.findall(LGRS_pattern,sys.argv[4])[0])
                        lonBand2,latBand2,e25k2,n25k2,E2,N2=list(re.findall(LGRS_pattern,sys.argv[5])[0])
                    except:
                        raise IndexError("Operation aborted: Number of inputs not correct."
                                    "\nInput LGRS fine gris (less than 25km):LonBandlatBande25kn25kEN"
                                    " LR:LonBandlatBande25kn25kEN"
                                    "\nor in a un condensed format:"
                                    "\nLGRS UL:LonBand latBand e25k n25k E N LR:LonBand latBand e25k n25k E N")
                else: 
                    raise IndexError("Operation aborted: Number of inputs not correct."
                                    "\nInput LGRS fine gris (less than 25km):LonBandlatBande25kn25kEN"
                                    " LR:LonBandlatBande25kn25kEN"
                                    "\nor in a un condensed format:"
                                    "\nLGRS UL:LonBand latBand e25k n25k E N LR:LonBand latBand e25k n25k E N")
                
                # convert to proper data formats
                try:
                    lonBand1 = int(lonBand1)
                    latBand1 = str(latBand1)
                    e25k1 = str(e25k1)
                    n25k1 = str(n25k1)
                    E1 = float(E1)
                    N1 = float(N1)

                    lonBand2 = int(lonBand2)
                    latBand2 = str(latBand2)
                    e25k2 = str(e25k2)
                    n25k2 = str(n25k2)
                    E2 = float(E2)
                    N2 = float(N2)
                except: 
                    raise IndexError("Operation aborted: LGRS Coordinates could not be"
                        " converted to proper format correctly.")
            else:
                raise IndexError("Operation aborted: LGRS grid size could not be determined"
                                ". Please select on of the following:\n{}".format(grid_format_str)) 
        
        # Lunar Grid Reference System in LPS regions
        elif form_in == "PolarLGRS":
            if Grid_Res not in grid_format[5:]: # 25km or greater

                if len(sys.argv)-1 == 4:
                    h = str(sys.argv[4])

                else: 
                    raise IndexError("Operation aborted: LGRS grid size could not be determined"
                            "Input for Polar LGRS coarse grid (>=25km): }"
                            "hemisphere \"S\" | \"N\"")

            elif Grid_Res in grid_format:
                if len(sys.argv)-1 == 13:

                    lonBand1 = sys.argv[4]
                    e25k1 = sys.argv[5]
                    n25k1 = sys.argv[6]
                    E1 = sys.argv[7]
                    N1 = sys.argv[8]

                    lonBand2 = sys.argv[9]
                    e25k2 = sys.argv[10]
                    n25k2 = sys.argv[11]
                    E2 = sys.argv[12]
                    N2 = sys.argv[13]

                elif len(sys.argv)-1 == 5:
                    try:
                        lonBand1,e25k1,n25k1,E1,N1 = list(re.findall(PolarLGRS_pattern,sys.argv[4])[0])
                        lonBand2,e25k2,n25k2,E2,N2 = list(re.findall(PolarLGRS_pattern,sys.argv[5])[0])
                    except:
                        raise IndexError("Operation aborted: Number of inputs not correct."
                        "\nInput for condensed PolarLGRS UL:LonBande25kn25kEN LR:LonBande25kn25kEN"
                        "\nor"
                        "\nPolarLGRS UL:LonBand e25k n25k E N LR:LonBand e25k n25k E N")

                else: 
                    raise IndexError("Operation aborted: Number of inputs not correct."
                        "\nInput for condensed PolarLGRS UL:LonBande25kn25kEN LR:LonBande25kn25kEN"
                        "\nor"
                        "\nPolarLGRS UL:LonBand e25k n25k E N LR:LonBand e25k n25k E N") 
                # convert to proper data formats
                try:
                    lonBand1 = str(lonBand1)
                    e25k1 = str(e25k1)
                    n25k1 = str(n25k1)
                    E1 = float(E1)
                    N1 = float(N1)

                    lonBand2 = str(lonBand2)
                    e25k2 = str(e25k2)
                    n25k2 = str(n25k2)
                    E2 = float(E2)
                    N2 = float(N2)   
                except: 
                    raise IndexError("Operation aborted: Polar LGRS Coordinates could not be"
                        " converted to proper format correctly.")


            else:
                raise IndexError("Operation aborted: PolarLGRS grid size could not be determined"
                                ". Please select on of the following:\n{}".format(grid_format_str)) 

        # Lunar Grid Reference System in ACC format in LTM regions
        elif form_in == "LGRS_ACC":

            if Grid_Res in grid_format[5:]: # any <= 25km or smaller
                if len(sys.argv)-1 == 19:

                    lonBand1 = int(sys.argv[4])
                    latBand1 = str(sys.argv[5])
                    e25k1 = str(sys.argv[6])
                    n25k1 = str(sys.argv[7])
                    e1k1 = str(sys.argv[8])
                    E1 = float(sys.argv[9])
                    n1k1 = str(sys.argv[10])
                    N1 = float(sys.argv[11])

                    lonBand2 = int(sys.argv[12])
                    latBand2 = str(sys.argv[13])
                    e25k2 = str(sys.argv[14])
                    n25k2 = str(sys.argv[15])
                    e1k2 = str(sys.argv[16])
                    E2 = float(sys.argv[17])
                    n1k2 = str(sys.argv[18])
                    N2 = float(sys.argv[19])

                elif len(sys.argv)-1 == 5:
                    try:
                        lonBand1,latBand1,e25k1,n25k1,e1k1,E1,n1k1,N1 = list(re.findall(LGRS_ACC_pattern,sys.argv[4])[0])
                        lonBand2,latBand2,e25k2,n25k2,e1k2,E2,n1k2,N2 = list(re.findall(LGRS_ACC_pattern,sys.argv[5])[0])

                    except:
                        raise IndexError("Operation aborted: Number of inputs not correct."
                            "\nInput for condensed LGRS ACC UL:LonBandlatBande25km25k1ekmE1nkmN LR:LonBandlatBande25km25km1ekmE1nkmN"
                            "\nor"
                            "\nLGRS ACC UL:LonBand latBand e25km 25km 1ekm E 1nkm N LR:LonBand latBand e25km 25km 1ekm E1 1nkm N")
                else: 
                    raise IndexError("Operation aborted: Number of inputs not correct."
                                    "\nInput for condensed LGRS ACC UL:LonBandlatBande25km25k1ekmE1nkmN LR:LonBandlatBande25km25km1ekmE1nkmN"
                                    "\nor"
                                    "\nLGRS ACC UL:LonBand latBand e25km 25km 1ekm E 1nkm N LR:LonBand latBand e25km 25km 1ekm E1 1nkm N")

                # convert to proper data formats
                try:
                    lonBand1 = int(lonBand1)
                    latBand1 = str(latBand1)
                    e25k1 = str(e25k1)
                    n25k1 = str(n25k1)
                    e1k1 = str(e1k1)
                    E1 = float(E1)
                    n1k1 = str(n1k1)
                    N1 = float(N1)

                    lonBand2 = int(lonBand2)
                    latBand2 = str(latBand2)
                    e25k2 = str(e25k2)
                    n25k2 = str(n25k2)
                    e1k2 = str(e1k2)
                    E2 = float(E2)
                    n1k2 = str(n1k2)
                    N2 = float(N2)
                except: 
                    raise IndexError("Operation aborted: LGRS_ACC Coordinates could not be"
                        " converted to proper format correctly.")

            else:
                raise IndexError("Operation aborted: Grid size for LGRS in ACC must be <=25km"
                                "Correct grid size could not be determined"
                                ". Please select on of the following:\n{}".format(grid_format[5:])) 
            
        # Lunar Grid Reference System in ACC format in LTM regions
        elif form_in == "PolarLGRS_ACC":
            if Grid_Res in grid_format[5:]: # any <= 25km or smaller
                if len(sys.argv)-1 == 17:
                    lonBand1 = sys.argv[4]
                    e25k1 = sys.argv[5]
                    n25k1 = sys.argv[6]
                    e1k1 = sys.argv[7]
                    E1 = sys.argv[8]
                    n1k1 = sys.argv[9]
                    N1 = sys.argv[10]

                    lonBand2 = sys.argv[11]
                    e25k2 = sys.argv[12]
                    n25k2 = sys.argv[13]
                    e1k2 = sys.argv[14]
                    E2 = sys.argv[15]
                    n1k2 = sys.argv[16]
                    N2 = sys.argv[17]

                elif len(sys.argv)-1 == 5:
                    try:
                        lonBand1,e25k1,n25k1,e1k1,E1,n1k1,N1 = list(re.findall(PolarLGRS_ACC_pattern,sys.argv[4])[0])
                        lonBand2,e25k2,n25k2,e1k2,E2,n1k2,N2 = list(re.findall(PolarLGRS_ACC_pattern,sys.argv[5])[0])
                    except: 
                        raise IndexError("Operation aborted: Number of inputs not correct."
                                "\nInput for condensed Polar LGRS ACC UL:LonBande25km25k1ekmE1nkmN LR:LonBandlatBande25km25km1ekmE1nkmN"
                                "\nor"
                                "\nPolar LGRS ACC UL:LonBand e25km 25km 1ekm E 1nkm N LR:LonBand e25km 25km 1ekm E1 1nkm N")
                else: 
                    raise IndexError("Operation aborted: Number of inputs not correct."
                                    "\nInput for condensed Polar LGRS ACC UL:LonBande25km25k1ekmE1nkmN LR:LonBandlatBande25km25km1ekmE1nkmN"
                                    "\nor"
                                    "\nPolar LGRS ACC UL:LonBand e25km 25km 1ekm E 1nkm N LR:LonBand e25km 25km 1ekm E1 1nkm N")
                
                # convert to proper data formats
                try:
                    lonBand1 = str(lonBand1)
                    e25k1 = str(e25k1)
                    n25k1 = str(n25k1)
                    e1k1 = str(e1k1)
                    E1 = float(E1)
                    n1k1 = str(n1k1)
                    N1 = float(N1)

                    lonBand2 = str(lonBand2)
                    e25k2 = str(e25k2)
                    n25k2 = str(n25k2)
                    e1k2 = str(e1k2)
                    E2 = float(E2)
                    n1k2 = str(n1k2)
                    N2 = float(N2)
                except: 
                    raise IndexError("Operation aborted: Polar LGRS ACC Coordinates could not be"
                        " converted to proper format correctly.")

            else:
                raise IndexError("Operation aborted: Grid size for LGRS in ACC must be <=25km"
                                "Correct grid size could not be determined"
                                ". Please select on of the following:\n{}".format(grid_format[5:])) 
            
        else:
            raise IndexError("Number of inputs not correct."
                            "Input 1. Coordinate import format"
                            "2. Coordinate output format Form_out "
                            "and the  corner points for a grid bounding"
                            "box: ul(X|Y) lr(X|Y)")

    except IndexError as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        print(exc_type, fname, exc_tb.tb_lineno)
        print(e)
        exit(1)

# ---------------------------------------------------------------------
# SECTION 3.3: Converting Input Coordinates to Common Grid Format. 
#           LPS or LTM format is chosen for simplicity. Note these are 
#           inverse conversions. Any truncation prior to this program will 
#           rounding of the original position. 
#           Program ends by checking to make sure that if two sets of 
#           Coordinates are provided, that that conform to having the 
#           upper left coordinate be coordinate 1 and the lower right 
#           to be coordinate 2.


    try: 
        if form_in == "LatLon":
            if Grid_Res == "global_all" or Grid_Res == "25km_all":
                z,h = None,None
            elif Grid_Res not in grid_format[5:]: # 25km or greater
                z,h,E,N = toLTM(cent_x,cent_y,zone=None,trunc_val=0,process_errors=True)
            
            elif Grid_Res in grid_format:
                z1,h1,ulx,uly = toLTM(ulx,uly,zone=None,trunc_val=0,process_errors=True)
                z2,h2,lrx,lry = toLTM(lrx,lry,zone=None,trunc_val=0,process_errors=True)
            else:
                raise ValueError("Operation aborted: Conversion Method from LatLon to common format failed.")
            
        elif form_in == "PolarLatLon":
            if Grid_Res not in grid_format[5:]: # 25km or greater
                h,E,N = toLPS(cent_x,cent_y,trunc_val=0,process_errors=True)

            elif Grid_Res in grid_format:
                h1,ulx,uly = toLPS(ulx,uly,trunc_val=0,process_errors=True)
                h2,lrx,lry = toLPS(lrx,lry,trunc_val=0,process_errors=True)
            else:
                raise ValueError("Operation aborted:  Conversion Method from PolarLatLon to common format failed.")

        elif form_in == "LGRS":
            if Grid_Res not in grid_format[5:]: # 25km or greater
                z = z
                h = h
            elif Grid_Res in grid_format:
                z1,h1,ulx,uly = LGRStoLTM(latBand1,lonBand1,e25k1,n25k1,E1,N1)
                z2,h2,lrx,lry = LGRStoLTM(latBand2,lonBand2,e25k2,n25k2,E2,N2)
            else:
                raise ValueError("Operation aborted: Conversion Method from LGRS to common format failed.")

        elif form_in == "PolarLGRS":
            if Grid_Res not in grid_format[5:]: # 25km or greater
                h = h
            elif Grid_Res in grid_format:
                h1,ulx,uly = LGRStoLPS(lonBand1,e25k1,n25k1,E1,N1,process_errors=True)
                h2,lrx,lry=LGRStoLPS(lonBand2,e25k2,n25k2,E2,N2,process_errors=True)
            else:
                raise ValueError("Operation aborted: Conversion Method from Polar LGRS to common format failed.")

        elif form_in == "LGRS_ACC":
            if Grid_Res in grid_format:
                z1,h1,ulx,uly = LGRS_ACCtoLTM(latBand1,lonBand1,e25k1,n25k1,e1k1,E1,n1k1,N1,ACC=False,process_errors=True)
                z2,h2,lrx,lry = LGRS_ACCtoLTM(latBand2,lonBand2,e25k2,n25k2,e1k2,E2,n1k2,N2,ACC=False,process_errors=True)
            else:
                raise ValueError("Operation aborted: Conversion Method from Polar LGRS_ACC to common format failed.")
            
        elif form_in == "PolarLGRS_ACC":
            if Grid_Res in grid_format[5:]:
                h1,ulx,uly = LGRS_ACCtoLPS(lonBand1,e25k1,n25k1,e1k1,E1,n1k1,N1,ACC=False,process_errors=True)
                h2,lrx,lry = LGRS_ACCtoLPS(lonBand2,e25k2,n25k2,e1k2,E2,n1k2,N2,ACC=False,process_errors=True)
            else:
                raise ValueError("Operation aborted: Conversion Method from Polar LGRS_ACC to common format failed.")
        elif form_in == "LPS":
            pass
        
        elif form_in == "LTM":
            if Grid_Res == "global_all" or Grid_Res == "25km_all":
                z,h = None,None
            else:
                pass
        else:
            raise ValueError("Operation aborted: Conversion Method To Common input format failed."
                            "Check input variables")

        # check Coordinates to make sure bounding box is formatted correctly
        if Grid_Res in grid_format[5:]:

            # test equatorial areas
            if ( ( form_in == "LTM") or \
                ( form_in == "LatLon") or \
                ( form_in == "LGRS") or \
                ( form_in == "LGRS_ACC" ) ):

                # convert to lat lon to test box area for all input formats
                # and across multiple zones
                ulx_test,uly_test=toLatLon(ulx,uly,z1,h1,process_errors=True)
                lrx_test,lry_test=toLatLon(lrx,lry,z2,h2,process_errors=True)

                # check to make sure grid area is specified correctly
                if ulx_test > lrx_test:
                    raise ValueError ("Grid Easting of ULX greater than LRX")
                elif ulx_test == lrx_test:
                    raise ValueError ("Grid ULX and LRX Eastings are equal, no area to plot")
                elif uly_test < lry_test:
                    raise ValueError ("Grid ULY northing less then LRY.")
                elif uly_test == lry_test:
                    raise ValueError ("Grid ULY and LRY are equal, no area to plot")
            # test polar areas
            elif ( ( form_in == "LPS") or \
                ( form_in == "PolarLatLon") or \
                ( form_in == "PolarLGRS") or \
                ( form_in == "PolarLGRS_ACC" ) ):
                
                # no multiplots for polar regions, conversion to lat lon 
                # not needed
                if ulx > lrx:
                    raise ValueError ("Grid Easting of ULX greater than LRX")
                elif ulx == lrx:
                    raise ValueError ("Grid ULX and LRX Eastings are equal, no area to plot")
                elif uly < lry:
                    raise ValueError ("Grid ULY northing less then LRY.")
                elif uly == lry:
                    raise ValueError ("Grid ULY and LRY are equal, no area to plot")

            else:
                raise ValueError ("Grid projection area could not be checked")

    # print faults
    except ValueError as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        print(exc_type, fname, exc_tb.tb_lineno)
        print(e)
        exit(1)

# ---------------------------------------------------------------------
# SECTION 3.4: Generate the Correct Grid to the desired resolution
#              as implemented 'border', 'global', 'global_all', '25km', 
#                           '25km_all', '1km', '100m', '10m', or '1m'. 
#              The grid generation is largely handled by other functions

    try:
        # This is the the procedure to process grids that are 25km 
        # resolution or higher
        if Grid_Res in grid_format[:5]: 
                
            if form_out == "LTM":
                GridX,GridY = generate_LTM_grid(Grid_Res,z,h)

            elif form_out == "LPS":
                GridX,GridY = generate_LPS_grid(Grid_Res,h)
            
            elif form_out == "LGRS":
                GridX,GridY,multigrid = generate_LGRS_grid(Grid_Res,z,h,
                                        LatLon_override=LatLon_override)

            elif form_out == "PolarLGRS":
                GridX,GridY = generate_PolarLGRS_grid(Grid_Res,h,
                                    LatLon_override=LatLon_override)

                pass
            else:
                raise ValueError("Operation failed: Large scale grid "
                        "> 25km could not be generated. Check input "
                        "Arguments")
            
        # plot grids that are < 25km resolution
        elif Grid_Res in grid_format: 
            
            # plot equatorial grid
            if form_out == "LGRS" or \
               form_out == "LGRS_ACC" or \
               form_out == "ACC":
                
                GridX,GridY,multigrid = generate_LGRS_grid(Grid_Res,
                                                           lonBand=None,
                                                           h=None,
                                                           z1=z1,
                                                           h1=h1,
                                                           ulx=ulx,
                                                           uly=uly,
                                                           z2=z2,
                                                           h2=h2,
                                                           lrx=lrx,
                                                           lry=lry,
                                                           multigrid=multigrid,
                                                           LatLon_override=LatLon_override)
                
            # plot polar grid
            elif form_out == "PolarLGRS" or \
                 form_out == "PolarLGRS_ACC" or \
                 form_out == "PolarACC": 
                
                GridX,GridY = generate_PolarLGRS_grid(Grid_Res,
                                                      h=None,
                                                      h1=h1,
                                                      ulx=ulx,
                                                      uly=uly,
                                                      h2=h2,
                                                      lrx=lrx,
                                                      lry=lry,
                                                      LatLon_override=LatLon_override)

            else: 
                raise ValueError("Grid Could not dimension. "
                                    "Size could not be determined.")
        else:
            raise ValueError("Operation failed: Grid could not be "
                             "generated.")
        
      # print faults
    except ValueError as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        print(exc_type, fname, exc_tb.tb_lineno)
        print(e)
        exit(1)  

# ---------------------------------------------------------------------
# SECTION 3.5: Convert grid Coordinates to correct output format.
#              The Projections are only LTM, LPS, or lat lon thus the 
#              grid coordinates will be in these formats. 


    try:    
        # The global format is processed in LatLon. 
        # This is the only grid resolution that needs to be converted
        # back to the proper common format. All other grids are 
        # generated in LTM or LPS.
        
        # reconvert global format to LTM or LPS
        if Grid_Res == "global":
            # process equatorial areas
            if ( (form_out == "LTM" ) or \
                 (form_out == "LGRS" ) ):
                    for i in range(0,GridX.shape[0]):
                        for j in range(0,GridX.shape[1]):

                            # break if polygon ends of if no polygon
                            if np.isnan(GridX[i,j]):
                                continue

                            # conversion to LTM
                            zone,hem,GridX[i,j],GridY[i,j] = toLTM(GridX[i,j],GridY[i,j],zone=None,
                                trunc_val=0,process_errors=False)
                        
            # convert polar areas
            elif ( ( form_out == "LPS" ) or \
                   ( form_out == "PolarLGRS" ) ):
                    for i in range(0,GridX.shape[0]):
                        for j in range(0,GridX.shape[1]):
                            
                            # break if polygons ends of if no polygon
                            if np.isnan(GridX[i,j]):
                                continue

                            # Convert to LPS stereographic Coordinates
                            hem,GridX[i,j],GridY[i,j] = toLPS(GridX[i,j],GridY[i,j],
                                trunc_val=0,eqs="Spherical",process_errors=False)
        
            else:
                raise ValueError ("Output format for grid in global "
                                "format incorrect.")
        
    except ValueError as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        print(exc_type, fname, exc_tb.tb_lineno)
        print(e)
        exit(1)  

# ---------------------------------------------------------------------
# SECTION 3.6: Determine grid cell names in specified output format
    # Determine grid size in meters and the coordinate 
    # length, The length is based on a truncated LGRS coordinate,
    # if ACC is used then the coordinate will be shorter and still
    # fit in the coordinate length
    

    try:
        if attribute:
            # estimated length of output coordinate 
            if ( ( Grid_Res == "25km" ) or \
                ( Grid_Res == "25km_all" ) ):
                let_len = 25
            elif Grid_Res == "1km":
                let_len = 9
            elif Grid_Res == "100m":
                let_len = 11
            elif Grid_Res == "10m":
                let_len = 13
            elif Grid_Res == "1m":
                let_len = 15
            else:
                let_len = 3

            # create binary data for exporting name
            GridName = np.chararray((GridX.shape[0],1),itemsize=let_len)

            # loop through all grid polygons
            for i in range(0,GridX.shape[0]):

                # continue to next polygon if row contained no values
                if np.isnan(GridX[i,0]):
                    continue
                else:
                    # calculate polygon center to determine grid name
                    CentX = np.nanmean(GridX[i,:])
                    CentY = np.nanmean(GridY[i,:])

                # at this point the grid center Coordinates are converted
                # to the desired output and written into the output 
                # naming array
                    
                # grid name for LTM output
                if form_out == "LTM": # additional conversion needed
                    if Grid_Res == "global_all":
                        z = Calc_LTM_zone(CentX)
                        h = Calc_hemisphere(CentY)
                    GridName[i,0] = "{}{}".format(z,h)

                # grid name for LPS output
                elif form_out == "LPS": # additional conversion needed
                    GridName[i,0] = "{}".format(h)

                # grid name for LGRS grid name 
                elif form_out == "LGRS":
                    # to determine names only, the center of grid formatted in
                    # latlon will need to be converted to LTM first
                    if multigrid or \
                       Grid_Res == "global_all" or \
                       Grid_Res == "25km_all" or \
                       LatLon_override:

                        # convert center of grid location to common format
                        z,h,CentX,CentY = toLTM(CentX,CentY,zone=None,
                                            trunc_val=0,process_errors=False)
                    
                    # if not multigrid, assign zone and hemisphere to a general 
                    # term that can be called. Z1 Z2 and H1 and H2 are the 
                    # same if this is called as multigrid are exported as latlon
                    elif Grid_Res in grid_format[5:]:
                        z = z1
                        h = h1

                    # calculate coordinate for naming
                    lonBand,latBand,e25k,n25k,E,N = toLGRS(CentX,CentY,z,h,
                                            trunc_val=1,process_errors=True)
                    
                    # create names for all resolutions
                    if Grid_Res == "global" or Grid_Res == "global_all" :
                        GridName[i,0] = "{}{}".format(lonBand,latBand)

                    elif Grid_Res == "25km_all" or Grid_Res == "25km":
                        GridName[i,0] = "{}{}{}{}".format(lonBand,latBand,e25k,n25k)

                    elif Grid_Res == "1km":
                        GridName[i,0] = "{}{}{}{}{}{}".format(lonBand,latBand,e25k,n25k,
                                                        E[:2],N[:2])
                    elif Grid_Res == "100m":
                        GridName[i,0] = "{}{}{}{}{}{}".format(lonBand,latBand,e25k,n25k,
                                                        E[:3],N[:3])
                    elif Grid_Res == "10m":
                        GridName[i,0] = "{}{}{}{}{}{}".format(lonBand,latBand,e25k,n25k,
                                                        E[:4],N[:4])
                    elif Grid_Res == "1m":
                        GridName[i,0] = "{}{}{}{}{}{}".format(lonBand,latBand,e25k,n25k,
                                                        E[:5],N[:5])  
                    else:
                        raise ValueError("Error: LGRS Grid coordinate names could not be "
                                         "Generated.")
                
                # process LGRS names for LGRS in ACC format
                elif form_out == "LGRS_ACC":
                    # and extra conversion is needed for grids generated in lat lon
                    if multigrid or \
                       Grid_Res == "global_all" or \
                       LatLon_override:
                        
                        z,h,CentX,CentY = toLTM(CentX,CentY,zone=None,
                                            trunc_val=0,process_errors=False)
                        
                    elif Grid_Res in grid_format[5:]:
                        z = z1
                        h = h1
                    
                    # calculate coordinate for naming
                    lonBand,latBand,e25k,n25k,e1k,E,n1k,N = toLGRS_ACC(CentX,CentY,z,h,
                                            trunc_val=1,ACC=False,process_errors=True)
                    
                    # write out grid names to the desired grid size
                    if Grid_Res == "global" or Grid_Res == "global_all" :
                        GridName[i,0] = "{}{}".format(lonBand,latBand)

                    elif Grid_Res == "25km_all" or Grid_Res == "25km":
                        GridName[i,0] = "{}{}{}{}".format(lonBand,latBand,e25k,n25k)

                    elif Grid_Res == "1km":
                        GridName[i,0] = "{}{}{}{}{}{}".format(lonBand,latBand,e25k,n25k,
                                                        e1k,n1k)
                    elif Grid_Res == "100m":
                        GridName[i,0] = "{}{}{}{}{}{}{}{}".format(lonBand,latBand,e25k,n25k,
                                                        e1k,E[:1],n1k,N[:1])
                    elif Grid_Res == "10m":
                        GridName[i,0] = "{}{}{}{}{}{}{}{}".format(lonBand,latBand,e25k,n25k,
                                                        e1k,E[:2],n1k,N[:2])
                    elif Grid_Res == "1m":
                        GridName[i,0] = "{}{}{}{}{}{}{}{}".format(lonBand,latBand,e25k,n25k,
                                                        e1k,E,n1k,N)
                    else:
                        raise ValueError("Error: LGRS ACC Grid coordinate names could not be "
                                        "Generated.")
                
                # For grids <25km resolution, write out names in ACC format
                # truncated to 10m without global location reference
                elif form_out == "ACC":
                    # convert latlong formatted grids back to common 
                    # format before naming 
                    if multigrid or LatLon_override:
                        z,h,CentX,CentY=toLTM(CentX,CentY,zone=None,
                                            trunc_val=0,process_errors=False)
                    
                    # Set projection zones, this should nt be called by any
                    # zone crossing in multiple zones as they are formatted 
                    # in LatLons to ensue naming. only grids in one 
                    # zone area up here
                    elif Grid_Res in grid_format[5:]:
                        z = z1
                        h = h1
                    
                    # convert coordinates for LGRS ACC format
                    e1k,E,n1k,N = toLGRS_ACC(CentX,CentY,z,h,
                                        trunc_val=10,ACC=True,process_errors=True)

                    if Grid_Res == "1km":
                        GridName[i,0] = "{}{}".format(e1k,n1k)
                    elif Grid_Res == "100m":
                        GridName[i,0] = "{}{}{}{}".format(e1k,E[:1],n1k,N[:1])
                    elif Grid_Res == "10m":
                        GridName[i,0] = "{}{}{}{}".format(e1k,E[:2],n1k,N[:2])
                    elif Grid_Res == "1m":
                        raise ValueError("Error: ACC Grid coordinate cannot be a 1m resolution. Plot "
                                        "LGRS_ACC grid instead.")
                    elif Grid_Res in grid_format:
                        raise ValueError("Error: ACC Grid coordinate specified incorrectly."
                                        "Check input grid resolution")
                    else:
                        raise ValueError("Error: ACC Grid coordinate names could not be "
                                        "generated.")
                    
                # names for polar LGRS 
                elif form_out == "PolarLGRS":
                    # extra conversion step for grids formatted in lat lon
                    if multigrid or \
                       Grid_Res == "global_all" or \
                       LatLon_override:

                        # latlon to LPS
                        h,CentX,CentY = toLPS(CentX,CentY,trunc_val=0,
                                            eqs="Spherical",process_errors=True)
                    
                    # grids formatted within a box need a common variable for 
                    # output. This has no affect as grids within only one zone 
                    # make it ot this check
                    elif Grid_Res in grid_format[5:]:
                        h = h1

                    # convert LPS coordinates to determine name
                    lonBand,e25k,n25k,E,N = toLGRS_polar(CentX,CentY,h,
                                            trunc_val=1,process_errors=True)
                    
                    # format name for output Coordinates
                    if Grid_Res == "global" or Grid_Res == "global_all" :
                        GridName[i,0] = "{}".format(lonBand)

                    elif Grid_Res == "25km_all" or Grid_Res == "25km":
                        GridName[i,0] = "{}{}{}".format(lonBand,e25k,n25k)

                    elif Grid_Res == "1km":
                        GridName[i,0] = "{}{}{}{}{}".format(lonBand,e25k,n25k,
                                                        E[:2],N[:2])
                    elif Grid_Res == "100m":
                        GridName[i,0] = "{}{}{}{}{}".format(lonBand,e25k,n25k,
                                                        E[:3],N[:3])
                    elif Grid_Res == "10m":
                        GridName[i,0]="{}{}{}{}{}".format(lonBand,e25k,n25k,
                                                        E[:4],N[:4])
                    elif Grid_Res == "1m":
                        GridName[i,0] = "{}{}{}{}{}".format(lonBand,e25k,n25k,
                                                        E[:5],N[:5])  
                    else:
                        raise ValueError("Error: PolarLGRS Grid coordinate names could not be "
                                        "generated.")

                # Polar LGRS coordinate names in ACC format without truncation
                elif form_out == "PolarLGRS_ACC":
                    
                    # extra processing step for grids formatted in lat long
                    if multigrid or \
                       Grid_Res == "global_all" or \
                       LatLon_override:
                        
                        # convert to LPS
                        h, CentX, CentY = toLPS(CentX, CentY,
                            trunc_val=0,
                            eqs="Spherical",
                            process_errors=False)


                    # assign projection hemisphere to common term for exporting
                    elif Grid_Res in grid_format[5:]:
                        h=h1

                    # Convert center of grid area to determine name
                    lonBand,e25k,n25k,e1k,E,n1k,N = toLGRS_polar_ACC(CentX,CentY,h,
                                            trunc_val=1,ACC=False,process_errors=True)
                    
                    # format output names for all grid resolutions
                    if Grid_Res == "global" or Grid_Res == "global_all" :
                        GridName[i,0] = "{}".format(lonBand)

                    elif Grid_Res == "25km_all" or Grid_Res == "25km":
                        GridName[i,0] = "{}{}{}".format(lonBand,e25k,n25k)

                    elif Grid_Res == "1km":
                        GridName[i,0] = "{}{}{}{}{}".format(lonBand,e25k,n25k,
                                                        e1k,n1k)
                    elif Grid_Res == "100m":
                        GridName[i,0] = "{}{}{}{}{}{}{}".format(lonBand,e25k,n25k,
                                                        e1k,E[:1],n1k,N[:1])
                    elif Grid_Res == "10m":
                        GridName[i,0] = "{}{}{}{}{}{}{}".format(lonBand,e25k,n25k,
                                                        e1k,E[:2],n1k,N[:2])
                    elif Grid_Res == "1m":
                        GridName[i,0] = "{}{}{}{}{}{}{}".format(lonBand,e25k,n25k,
                                                        e1k,E,n1k,N)
                    else:
                        raise ValueError("Error: PolarLGRS ACC Grid coordinate names could not be "
                                        "Generated.")
                    
                # Polar LGRS coordinate names in ACC format with truncation
                elif form_out == "PolarACC":

                    # grids covering multiple zones are in lat lon and need to 
                    # be converted to LTM before names can be determined
                    if multigrid or LatLon_override:
                        
                        # convert to LTM
                        h, CentX, CentY = toLPS(CentX, CentY,
                            trunc_val=0,
                            eqs="Spherical",
                            process_errors=False)

                    
                    # For grids in single zone assign output hemisphere
                    elif Grid_Res in grid_format[5:]:
                        h = h1
                    
                    # determine PolarACC Grid Names
                    e1k,E,n1k,N = toLGRS_polar_ACC(CentX,CentY,h,
                                    trunc_val=10,ACC=True,process_errors=True)

                    # format output names
                    if Grid_Res == "1km":
                        GridName[i,0] = "{}{}".format(e1k,n1k)
                    elif Grid_Res == "100m":
                        GridName[i,0] = "{}{}{}{}".format(e1k,E[:1],n1k,N[:1])
                    elif Grid_Res == "10m":
                        GridName[i,0] = "{}{}{}{}".format(e1k,E[:2],n1k,N[:2])
                    elif Grid_Res == "1m":
                        raise ValueError("Error: Polar ACC Grid coordinate be a 1m resolution. Plot "
                                        "LGRS_ACC grid instead.")
                    elif Grid_Res in grid_format:
                        raise ValueError("Error: PolarACC Grid coordinate specified incorrectly."
                                        "Check input grid resolution")
                    else:
                        raise ValueError("Error: ACC Grid coordinate names could not be "
                                        "generated.")
        else:
            GridName = None
        
    # report errors if discovered
    except ValueError as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        print(exc_type, fname, exc_tb.tb_lineno)
        print(e)
        exit(1)

# ---------------------------------------------------------------------
# SECTION 3.7: Assign output WKT1 Projection string for writing file 
        

    try:
        # Export in LatLon Coordinates 
        if ( ( multigrid ) or 
             ( LatLon_override ) or \
             ( Grid_Res == "border" ) or \
             ( Grid_Res == "global_all" ) or \
             ( Grid_Res == "25km_all" ) ):
            WKT1 = Display_geocentric_degrees

        # export Equatorial Grids
        else:
            if ( ( form_out == "LTM" ) or \
                ( form_out == "LGRS" ) or \
                ( form_out == "LGRS_ACC" ) or \
                ( form_out == "ACC" ) ):
                
                # Step used to assign correct zone
                if Grid_Res in grid_format[5:]:
                    z = z1
                    h = h1
                    
                # determine LTM zone central meridian
                lam0 = Calc_LTM_prime_meridian(z,"deg")    

                # Depending on hemisphere assign correct projection 
                # to data           
                if h == "N":
                    WKT1 = N_LTM_projection_meters.format(lam0)
                elif h == "S":
                    WKT1 = S_LTM_projection_meters.format(lam0)

            # export polar grids
            elif ( ( form_out == "LPS" )  or \
                ( form_out == "PolarLGRS" ) or \
                ( form_out == "PolarLGRS_ACC" ) or \
                ( form_out == "PolarACC" ) ):
                
                # step used to determine hemisphere
                if Grid_Res in grid_format[5:]:
                    h = h1

                # assign projection based on the hemispheric 
                # location
                if h == "N":
                    WKT1 = N_LPS_projection_meters
                elif h == "S":
                    WKT1 = S_LPS_projection_meters
            else:
                raise ValueError ("Map output projection could not be determined")
            
    except ValueError as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        print(exc_type, fname, exc_tb.tb_lineno)
        print(e)
        exit(1)

# ---------------------------------------------------------------------
# SECTION 3.8:  Write grid to ESRI shape file and format output names


    try:
        # This is the the procedure to process grids that are 25km 
        # resolution or higher

        # format shape file output names for equatorial region
        if ( ( form_out == "LTM" ) or \
             ( form_out == "LGRS" ) or \
             ( form_out == "LGRS" ) or \
             ( form_out == "LGRS_ACC" ) or \
             ( form_out == "ACC" ) ):
            
            # global grid names
            if Grid_Res == "global_all" or Grid_Res == "25km_all":
                f_o = "{}_{}_Grid".format(form_out,Grid_Res)
            else:
                # multigrid names for grid passing though multiple zones
                if multigrid:
                    f_o = "{}_{}_{}{}_MultiGrid".format(form_out,Grid_Res,z,h)
                
                # Output name for grids in LTM equatorial regions
                else:
                    f_o = "{}_{}_{}{}_Grid".format(form_out,Grid_Res,z,h)
        
        # output names for polar regions
        else:
            f_o = "{}_{}_{}_Grid".format(form_out,Grid_Res,h)

        # write out to shape file for single polygon
        if Grid_Res == "border" or Grid_Res == "global":
            
            # Export LTM and LPS single polygons
            if form_out == "LTM" or form_out == "LPS":
                Xport_polygon(GridX,
                              GridY,
                              GridName,
                              f_o,
                              WKT1,
                              attribute,
                              output_file_form,
                              overwrite)
                
            # only LTM and LPS have single polygon grids
            # all other global grids are processed in
            # multipolygon geometries
            else:
                Xport_multipolygons(GridX,
                                    GridY,
                                    GridName,
                                    f_o,
                                    WKT1,
                                    attribute,
                                    output_file_form,
                                    overwrite)
                
        # export all other grids as a multipolygon
        elif Grid_Res in grid_format:
            Xport_multipolygons(GridX,
                                GridY,
                                GridName,
                                f_o,
                                WKT1,
                                attribute,
                                output_file_form,
                                overwrite)
        else:
            raise ValueError ("Grid could not be exported. "
                              "Please check function input values ")
      
        # print faults
    except ValueError as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        print(exc_type, fname, exc_tb.tb_lineno)
        print(e)
        exit(1)

# ---------------------------------------------------------------------
# Print program timing and elapsed time


    if info:
        EndTime = datetime.datetime.now()
        print('\n', ProgName, 'execution time')
        print('    Start Time:', StartTime)
        print('    End Time:  ', EndTime)
        print('    Run Time:  ', EndTime - StartTime)


# =====================================================================
# Run main program
# determine input argument format

# main program control and operations
multigrid = False  # process grid across LTM area
attribute = False   # write out shape file attribute information
overwrite = True   # delete and requite file if it exits 
output_file_form = "shp"  # set the type output file. 
                           # shp for esri shapefile
                           # gpkg for a geopack file
info = True  # print out program information and processing time
LatLon_override = False # force the grids to output in lat lons
                        # only applies to LGRS and LGRS in ACC format
                        # NOTE: these projections must still be used
                        # in LPS or LTM to ensure correct spatial
                        # dimensions.

# initialize program information
if info:
    StartTime = datetime.datetime.now()
    print('Program:', ProgName)


try:
    # read in initial arguments
    initialize_LGRS_grid_generation_globals()
    form_in = sys.argv[1]
    form_out = sys.argv[2]
    grid_res = sys.argv[3]

# raise problem is no basic information is provided to the 
# program
except IndexError:
    form_in = None
    form_out = None
    grid_res = None
    raise ValueError("Input, output, or grid resolution specified "
        "incorrectly.\n Select on of the following:\n"
        "I/O:\n{}\n\nGrid Res:\n{}".format(coordinate_format_str,
                                           grid_format_str))

# run the grid generation program if enough arguments are provided
try:
    if ( (form_in in coordinate_format) and \
         ( (form_out in coordinate_format) or \
           (form_out == "ACC") or \
           (form_out == "PolarACC") ) and \
         ( grid_res in grid_format ) ):
        
        # call main function to generate grids
        main(form_in,form_out,grid_res,
                              multigrid,
                              attribute,
                              overwrite,
                              output_file_form,
                              info,
                              LatLon_override)

    else:
        raise ValueError("Input, output, or grid resolution specified "
                "incorrectly.\n Select on of the following:\n"
                "I/O:\n{}\n\nGrid Res:\n{}".format(coordinate_format_str,
                                                   grid_format_str))
    
except ValueError as e:

    exc_type, exc_obj, exc_tb = sys.exc_info()
    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
    print(exc_type, fname, exc_tb.tb_lineno)
    print(e)

# =====================================================================


